import pandas as pd
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord

class Catalog(object):
    """ Base class for results from catalog queries
    """
    _distance_column = '_r'

    def __init__(self, query):
        self.query = query
        self._table = None
        self._query_coords = None
        self._coords = None
        self._empty = False

    def __repr__(self):
        return '{0}({1})'.format(type(self), repr(self.query))

    def __str__(self):
        return '{} Query of {}'.format(self.name, self.query)

    @property
    def coords(self):
        if self._coords is None:
            self._run_query()
        return self._coords


    @property
    def query_coords(self):
        if self._query_coords is None:
            q = self.query
            dt = (q.epoch - self.epoch)*u.yr
            ra = q.ra*u.deg - dt*q.pmra*u.mas/u.yr
            dec = q.dec*u.deg - dt*q.pmdec*u.mas/u.yr
            self._query_coords = SkyCoord(ra, dec)
        return self._query_coords

    @property
    def table(self):
        if self._table is None:
            self._run_query()
        return self._table

    @property
    def df(self):
        return pd.DataFrame(np.array(self.table))

    @property
    def closest(self):
        df = self.df.sort_values(by=self._distance_column)
        return df.iloc[0]

    @property
    def brightest(self):
        band = list(self.bands.keys())[0]
        df = self.df.sort_values(by=band)
        return df.iloc[0]

    def get_id(self, brightest=False):
        if brightest:
            row = self.brightest
        else:
            row = self.closest

        return row[self.id_column]

    def get_photometry(self, brightest=False,
                    min_unc=0.02, convert=True):
        """Returns dictionary of photometry of closest match

        unless brightest is True, in which case the brightest match.
        """
        if brightest:
            row = self.brightest
        else:
            row = self.closest

        if not hasattr(self, 'conversions'):
            convert = False

        if convert:
            bands = self.conversions
        else:
            bands = self.bands.keys()

        d = {}
        for b in bands:
            if convert:
                key = b
                mag, dmag = getattr(self, b)(brightest=brightest)
            else:
                key = self.bands[b]
                mag, dmag = row[b], row['e_{}'.format(b)]

            d[key] = mag, max(dmag, min_unc)
        return d



import os, glob
import numpy as np
import tempfile
import tables as tb


from isochrones.dartmouth import Dartmouth_Isochrone
from isochrones.mist import MIST_Isochrone
from isochrones import StarModel
from isochrones.starfit import starfit

mnest = True
try:
    import pymultinest
except:
    import logging
    logging.warning('No PyMultiNest; fits will use emcee')
    mnest = False

chainsdir = tempfile.gettempdir()

props = dict(Teff=(5800, 100), logg=(4.5, 0.1),
             B=(5.7,0.05), V=(5.0, 0.05))

def test_fitting():
    mod_dar = _check_fitting(StarModel(Dartmouth_Isochrone, **props))
    mod_mist = _check_fitting(StarModel(MIST_Isochrone, **props))

    _check_saving(mod_dar)
    _check_saving(mod_mist)

def test_starfit():
    rootdir = os.path.abspath(os.path.join(os.path.dirname(__file__)))
    testdir = os.path.join(rootdir, 'star1')
    if mnest:
        basename = '{}/{}-'.format(chainsdir,np.random.randint(1000000))
        kwargs = dict(n_live_points=20, max_iter=100,basename=basename,
                      verbose=False)
    else:
        kwargs = dict(nburn=20, niter=20, ninitial=10)

    mod, _ = starfit(testdir, overwrite=True, use_emcee=not mnest,
                     no_plots=True, **kwargs)

    mod.samples

    if mnest:
        files = glob.glob('{}*'.format(basename))
        for f in files:
            os.remove(f)

###############

def _check_saving(mod):
    filename = os.path.join(chainsdir, '{}.h5'.format(np.random.randint(1000000)))
    mod.save_hdf(filename)
    assert len(tb.file._open_files.get_handlers_by_name(filename)) == 0

    newmod = StarModel.load_hdf(filename)
    assert len(tb.file._open_files.get_handlers_by_name(filename)) == 0

    assert np.allclose(mod.samples, newmod.samples)
    assert mod.ic.bands == newmod.ic.bands


    os.remove(filename)

def _check_fitting(mod):
    _fit_emcee(mod)
    if mnest:
        _fit_mnest(mod)
    return mod

def _fit_mnest(mod):
    basename = '{}/{}-'.format(chainsdir,np.random.randint(1000000))
    mod.fit_multinest(n_live_points=5, max_iter=50,basename=basename,
            verbose=False)
    foo = mod.mnest_analyzer
    files = glob.glob('{}*'.format(basename))
    for f in files:
        os.remove(f)

def _fit_emcee(mod):
    mod.use_emcee = True
    mod.fit_mcmc(nburn=20, niter=10, ninitial=10)
    mod.samples



import os
import pickle

from ..isochrone import Isochrone
from ..config import ISOCHRONES
from .grid import YAPSIModelGrid

class YAPSI_Isochrone(Isochrone):
    
    name = 'dartmouth'
    default_bands = YAPSIModelGrid.default_bands
    tri_file = os.path.join(ISOCHRONES, 'yapsi.tri')
    
    def __init__(self, **kwargs):
        df = YAPSIModelGrid().df

        with open(self.tri_file, 'rb') as f:
            tri = pickle.load(f)
        
        mags = {b:df[b].values for b in self.default_bands}

        Isochrone.__init__(self,df['mass'].values, df['age'].values,
                           df['feh'].values,df['mass'].values, df['logL'].values,
                           10**df['logTeff'].values,df['logg'].values,mags,tri=tri, 
                           **kwargs)

        
        


from isochrones.query import Query, TwoMASS, WISE, Tycho2, Gaia

import sys
if sys.version_info < (3,):
    def b(x):
        return x
else:
    import codecs
    def b(x):
        return codecs.latin_1_encode(x)[0]

def test_queries(ra=45.03433035439128, dec=0.23539164875137225,
                    pmra=43.75231341609215, pmdec=-7.6419899883511482,
                    epoch=2015.):
    """Testing with first entry from Gaia DR1 TGAS table
    """
    q = Query(ra, dec, pmra=pmra, pmdec=pmdec, epoch=epoch)

    tm = TwoMASS(q)
    w = WISE(q)
    tyc = Tycho2(q)
    gaia = Gaia(q)

    assert tm.get_id() == b('03000819+0014074') #force byte literal b/c that's what gets returned
    assert w.get_id() == b('J030008.22+001407.4')
    assert tyc.get_id() == '55-256-1'
    assert gaia.get_id() == 7632157690368



from __future__ import print_function, division
import os,os.path, glob
import logging
import re
import json

from configobj import ConfigObj

try:
    import numpy as np
except ImportError:
    np = None

try:
    import pandas as pd
except ImportError:
    pd = None
    
import numpy.random as rand
import scipy.optimize

try:
    import emcee
except ImportError:
    emcee = None

try:
    from plotutils.plotutils import setfig
except ImportError:
    setfig = None
    
try:
    import matplotlib.pyplot as plt
    from scipy.stats import gaussian_kde
except ImportError:
    plt = None
    gaussian_kde = None

try:
    import triangle
except ImportError:
    triangle = None

mnest_available = True
try:
    import pymultinest
except ImportError:
    logging.warning('PyMultiNest not available; only emcee fits will be possible.')
    pymultinest = None
    mnest_available = False


from .extinction import EXTINCTION
from .passbands import WEFF

class StarModel(object):
    """An object to represent a star, with observed properties, modeled by an Isochrone

    This is used to fit a physical stellar model to observed
    quantities, e.g. spectroscopic or photometric, based on
    an :class:`Isochrone`.  Parallax (in miliarcseconds) is
    also accepted as an observed quantity.

    Note that by default a local metallicity prior, based on SDSS data,
    will be used when :func:`StarModel.fit` is called.

    :param ic: 
        :class:`Isochrone` object used to model star.

    :param maxAV: (optional)
        Maximum allowed extinction (i.e. the extinction @ infinity in direction of star).  Default is 1.

    :param max_distance: (optional)
        Maximum allowed distance (pc).  Default is 3000.
    
    :param use_emcee: (optional)
        If set to true, then sampling done with emcee rather than MultiNest.
        (not recommended unless you have very precise spectroscopic properties).

    :param **kwargs:
        Keyword arguments must be properties of given isochrone, e.g., logg,
        feh, Teff, and/or magnitudes.  The values represent measurements of
        the star, and must be in (value,error) format. All such keyword
        arguments will be held in ``self.properties``.  ``parallax`` is
        also a valid property, and should be provided in miliarcseconds.
        
    """
    def __init__(self,ic,maxAV=1,max_distance=3000,
                 use_emcee=False, 
                 min_logg=None, name='',
                 **kwargs):
        self._ic = ic
        self.properties = kwargs
        self.max_distance = max_distance
        self.maxAV = maxAV

        self.name = name

        self._samples = None
        self._mnest_samples = None
        self.use_emcee = use_emcee
        if not mnest_available:
            logging.warning('MultiNest not available; use_emcee being set to True')
            self.use_emcee = True

        self.min_logg = min_logg
            
        self.n_params = 5 #mass, feh, age, distance, AV
        
        self._props_cleaned = False
        self._mnest_basename = None


    @property
    def ic(self):
        if type(self._ic)==type:
            self._ic = self._ic()
        return self._ic

    @classmethod
    def from_ini(cls, ic, folder='.', ini_file='star.ini'):
        """
        Initialize a StarModel from a .ini file

        File should contain all arguments with which to initialize
        StarModel.  
        """
        if not os.path.isabs(ini_file):
            ini_file = os.path.join(folder,ini_file)

        config = ConfigObj(ini_file)
        kwargs = {}
        for kw in config.keys():
            try:
                kwargs[kw] = float(config[kw])
            except:
                kwargs[kw] = (float(config[kw][0]), float(config[kw][1]))

        return cls(ic, **kwargs)
    
    @property
    def mags(self):
        d = {}
        for prop,vals in self.properties.items():
            if prop in self.ic.bands:
                try:
                    val,err = vals
                except TypeError:
                    val = vals
                d[prop] = val
        return d

    @property
    def mag_errs(self):
        d = {}
        for prop,vals in self.properties.items():
            if prop in self.ic.bands:
                try:
                    val,err = vals
                except TypeError:
                    continue
                d[prop] = err
        return d
    
    @property
    def Teff(self):
        if 'Teff' in self.properties:
            return self.properties['Teff']

    @property
    def feh(self):
        if 'feh' in self.properties:
            return self.properties['feh']

    @property
    def logg(self):
        if 'logg' in self.properties:
            return self.properties['logg']


    def _clean_props(self):
        """
        Makes sure all properties are legit for isochrone.

        Not done in __init__ in order to save speed on loading.
        """
        remove = []
        for p in self.properties.keys():
            if not hasattr(self.ic, p) and \
              p not in self.ic.bands and p not in ['parallax','feh','age','mass_B','mass_C'] and \
              not re.search('delta_',p):
                remove.append(p)

        for p in remove:
            del self.properties[p]

        if len(remove) > 0:
            logging.warning('Properties removed from Model because ' +
                            'not present in {}: {}'.format(type(self.ic),remove))

        remove = []
        for p in self.properties.keys():
            try:
                val = self.properties[p][0]
                if not np.isfinite(val):
                    remove.append(p)
            except:
                pass

        for p in remove:
            del self.properties[p]

        if len(remove) > 0:
            logging.warning('Properties removed from Model because ' +
                            'value is nan or inf: {}'.format(remove))
        


        self._props_cleaned = True
    
    def add_props(self,**kwargs):
        """
        Adds observable properties to ``self.properties``.
        
        """
        for kw,val in kwargs.iteritems():
            self.properties[kw] = val

    def remove_props(self,*args):
        """
        Removes desired properties from ``self.properties``.
        
        """
        for arg in args:
            if arg in self.properties:
                del self.properties[arg]
    
    @property
    def fit_for_distance(self):
        """
        ``True`` if any of the properties are apparent magnitudes.
        
        """
        for prop in self.properties.keys():
            if prop in self.ic.bands:
                return True
        return False
            
    def loglike(self, *args, **kwargs):
        """For backwards compatibility
        """
        return lnpost(*args, **kwargs)

    def lnlike(self, p):
        """Log-likelihood of model at given parameters

        
        :param p: 
            mass, log10(age), feh, [distance, A_V (extinction)].
            Final two should only be provided if ``self.fit_for_distance``
            is ``True``; that is, apparent magnitudes are provided.
            
        :return:
           log-likelihood.  Will be -np.inf if values out of range.
        
        """

        if not self._props_cleaned:
            self._clean_props()
            
        if not self.use_emcee:
            fit_for_distance = True
            mass, age, feh, dist, AV = (p[0], p[1], p[2], p[3], p[4])
        else:
            if len(p)==5:
                fit_for_distance = True
                mass,age,feh,dist,AV = p
            elif len(p)==3:
                fit_for_distance = False
                mass,age,feh = p

        if mass < self.ic.minmass or mass > self.ic.maxmass \
           or age < self.ic.minage or age > self.ic.maxage \
           or feh < self.ic.minfeh or feh > self.ic.maxfeh:
            return -np.inf
        if fit_for_distance:
            if dist < 0 or AV < 0 or dist > self.max_distance:
                return -np.inf
            if AV > self.maxAV:
                return -np.inf

        if self.min_logg is not None:
            logg = self.ic.logg(mass,age,feh)
            if logg < self.min_logg:
                return -np.inf

        logl = 0
        for prop in self.properties.keys():
            try:
                val,err = self.properties[prop]
            except TypeError:
                #property not appropriate for fitting (e.g. no error provided)
                continue

            if prop in self.ic.bands:
                if not fit_for_distance:
                    raise ValueError('must fit for mass, age, feh, dist, A_V if apparent magnitudes provided.')
                mod = self.ic.mag[prop](mass,age,feh) + 5*np.log10(dist) - 5
                A = AV*EXTINCTION[prop]
                mod += A
            elif re.search('delta_',prop):
                continue
            elif prop=='feh':
                mod = feh
            elif prop=='parallax':
                mod = 1./dist * 1000
            else:
                mod = getattr(self.ic,prop)(mass,age,feh)

            logl += -(val-mod)**2/(2*err**2) + np.log(1/(err*np.sqrt(2*np.pi)))

        if np.isnan(logl):
            logl = -np.inf

        return logl

    def lnprior(self, mass, age, feh,
                distance=None, AV=None, 
                use_local_fehprior=True):
        """
        log-prior for model parameters
        
        """
        mass_prior = salpeter_prior(mass)
        if mass_prior==0:
            mass_lnprior = -np.inf
        else:
            mass_lnprior = np.log(mass_prior)

        if np.isnan(mass_lnprior):
            logging.warning('mass prior is nan at {}'.format(mass))

        age_lnprior = np.log(age * (2/(self.ic.maxage**2-self.ic.minage**2)))
        if np.isnan(age_lnprior):
            logging.warning('age prior is nan at {}'.format(age))


        if use_local_fehprior:
            fehdist = local_fehdist(feh)
        else:
            fehdist = 1/(self.ic.maxfeh - self.ic.minfeh)
        feh_lnprior = np.log(fehdist)
        if np.isnan(feh_lnprior):
            logging.warning('feh prior is nan at {}'.format(feh))


        if distance is not None:
            if distance <= 0:
                distance_lnprior = -np.inf
            else:
                distance_lnprior = np.log(3/self.max_distance**3 * distance**2)
        else:
            distance_lnprior = 0
        if np.isnan(distance_lnprior):
            logging.warning('distance prior is nan at {}'.format(distance))


        if AV is not None:
            AV_lnprior = np.log(1/self.maxAV)
        else:
            AV_lnprior = 0
        if np.isnan(AV_lnprior):
            logging.warning('AV prior is nan at {}'.format(AV))
            

        lnprior = (mass_lnprior + age_lnprior + feh_lnprior + 
                distance_lnprior + AV_lnprior)

        return lnprior

    def lnpost(self, p, use_local_fehprior=True):
        """
        log-posterior of model at given parameters
        """
        if not self.use_emcee:
            mass, age, feh, dist, AV = (p[0], p[1], p[2], p[3], p[4])
        else:
            if len(p)==5:
                fit_for_distance = True
                mass,age,feh,dist,AV = p
            elif len(p)==3:
                fit_for_distance = False
                mass,age,feh = p
                dist = None
                AV = None
            
        return (self.lnlike(p) + 
                self.lnprior(mass, age, feh, dist, AV,
                             use_local_fehprior=use_local_fehprior))



    def maxlike(self,nseeds=50):
        """Returns the best-fit parameters, choosing the best of multiple starting guesses

        :param nseeds: (optional)
            Number of starting guesses, uniformly distributed throughout
            allowed ranges.  Default=50.

        :return:
            list of best-fit parameters: ``[m,age,feh,[distance,A_V]]``.
            Note that distance and A_V values will be meaningless unless
            magnitudes are present in ``self.properties``.
        
        """
        m0,age0,feh0 = self.ic.random_points(nseeds)
        d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nseeds))
        AV0 = rand.uniform(0,self.maxAV,size=nseeds)

        

        costs = np.zeros(nseeds)

        if self.fit_for_distance:
            pfits = np.zeros((nseeds,5))
        else:
            pfits = np.zeros((nseeds,3))
            
        def fn(p): #fmin is a function *minimizer*
            return -1*self.lnpost(p)
        
        for i,m,age,feh,d,AV in zip(range(nseeds),
                                    m0,age0,feh0,d0,AV0):
                if self.fit_for_distance:
                    pfit = scipy.optimize.fmin(fn,[m,age,feh,d,AV],disp=False)
                else:
                    pfit = scipy.optimize.fmin(fn,[m,age,feh],disp=False)
                pfits[i,:] = pfit
                costs[i] = self.lnpost(pfit)

        return pfits[np.argmax(costs),:]

    def mnest_prior(self, cube, ndim, nparams):
        """
        Transforms unit cube into parameter cube.

        Parameters if running multinest must be mass, age, feh, distance, AV.
        """
        cube[0] = (self.ic.maxmass - self.ic.minmass)*cube[0] + self.ic.minmass
        cube[1] = (self.ic.maxage - self.ic.minage)*cube[1] + self.ic.minage
        cube[2] = (self.ic.maxfeh - self.ic.minfeh)*cube[2] + self.ic.minfeh
        cube[3] = cube[3]*self.max_distance
        cube[4] = cube[4]*self.maxAV
    
    def mnest_loglike(self, cube, ndim, nparams):
        """loglikelihood function for multinest
        """
        return self.lnpost(cube)

    def fit(self, **kwargs):
        """
        Wrapper for either :func:`fit_multinest` or :func:`fit_mcmc`.

        Default will be to use MultiNest; set `use_emcee` keyword to `True` 
        if you want to use MCMC, or just call :func:`fit_mcmc` directly.
        """
        if self.use_emcee:
            if 'basename' in kwargs:
                del kwargs['basename']
            if 'verbose' in kwargs:
                del kwargs['verbose']
            if 'overwrite' in kwargs:
                del kwargs['overwrite']
            self.fit_mcmc(**kwargs)
        else:
            self.fit_multinest(**kwargs)

    def fit_multinest(self, n_live_points=1000, basename='chains/single-',
                      verbose=True, refit=False, overwrite=False,
                      **kwargs):
        """
        Fits model using MultiNest, via pymultinest.  

        :param n_live_points:
            Number of live points to use for MultiNest fit.

        :param basename:
            Where the MulitNest-generated files will live.  
            By default this will be in a folder named `chains`
            in the current working directory.  Calling this 
            will define a `_mnest_basename` attribute for 
            this object.

        :param verbose:
            Whether you want MultiNest to talk to you.

        :param refit, overwrite:
            Set either of these to true if you want to 
            delete the MultiNest files associated with the
            given basename and start over.

        :param **kwargs:
            Additional keyword arguments will be passed to 
            :func:`pymultinest.run`.

        """
        folder = os.path.abspath(os.path.dirname(basename))
        if not os.path.exists(folder):
            os.makedirs(folder)

        #If previous fit exists, see if it's using the same
        # observed properties
        prop_nomatch = False
        propfile = '{}properties.json'.format(basename)
        if os.path.exists(propfile):
            with open(propfile) as f:
                props = json.load(f)
            if set(props.keys()) != set(self.properties.keys()):
                prop_nomatch = True
            else:
                for k,v in props.items():
                    if np.size(v)==2:
                        if not self.properties[k][0] == v[0] and \
                                self.properties[k][1] == v[1]:
                            props_nomatch = True
                    else:
                        if not self.properties[k] == v:
                            props_nomatch = True

        if prop_nomatch and not overwrite:
            raise ValueError('Properties not same as saved chains ' +
                            '(basename {}*). '.format(basename) +
                            'Use overwrite=True to fit.')


        if refit or overwrite:
            files = glob.glob('{}*'.format(basename))
            [os.remove(f) for f in files]

        self._mnest_basename = basename

        pymultinest.run(self.mnest_loglike, self.mnest_prior, self.n_params,
                        n_live_points=n_live_points, outputfiles_basename=basename,
                        verbose=verbose,
                        **kwargs)

        with open(propfile, 'w') as f:
            json.dump(self.properties, f, indent=2)

        self._make_samples()

    @property
    def mnest_analyzer(self):
        """
        PyMultiNest Analyzer object associated with fit.  

        See PyMultiNest documentation for more.
        """
        return pymultinest.Analyzer(self.n_params, self._mnest_basename)

    @property
    def evidence(self):
        """
        Log(evidence) from multinest fit
        """
        s = self.mnest_analyzer.get_stats()
        return (s['global evidence'],s['global evidence error'])

    def fit_mcmc(self,nwalkers=300,nburn=200,niter=100,
                 p0=None,initial_burn=None,
                 ninitial=100, loglike_kwargs=None,
                 **kwargs):
        """Fits stellar model using MCMC.

        :param nwalkers: (optional)
            Number of walkers to pass to :class:`emcee.EnsembleSampler`.
            Default is 200.

        :param nburn: (optional)
            Number of iterations for "burn-in."  Default is 100.

        :param niter: (optional)
            Number of for-keeps iterations for MCMC chain.
            Default is 200.

        :param p0: (optional)
            Initial parameters for emcee.  If not provided, then chains
            will behave according to whether inital_burn is set.

        :param initial_burn: (optional)
            If `True`, then initialize walkers first with a random initialization,
            then cull the walkers, keeping only those with > 15% acceptance
            rate, then reinitialize sampling.  If `False`, then just do
            normal burn-in.  Default is `None`, which will be set to `True` if
            fitting for distance (i.e., if there are apparent magnitudes as
            properties of the model), and `False` if not.
            
        :param ninitial: (optional)
            Number of iterations to test walkers for acceptance rate before
            re-initializing.

        :param loglike_args:
            Any arguments to pass to :func:`StarModel.loglike`, such 
            as what priors to use.

        :param **kwargs:
            Additional keyword arguments passed to :class:`emcee.EnsembleSampler`
            constructor.
            
        :return:
            :class:`emcee.EnsembleSampler` object.
            
        """

        #clear any saved _samples
        if self._samples is not None:
            self._samples = None
            

        if self.fit_for_distance:
            npars = 5
            if initial_burn is None:
                initial_burn = True
        else:
            if initial_burn is None:
                initial_burn = False
            npars = 3

        if p0 is None:
            m0,age0,feh0 = self.ic.random_points(nwalkers)
            d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nwalkers))
            AV0 = rand.uniform(0,self.maxAV,size=nwalkers)
            if self.fit_for_distance:
                p0 = np.array([m0,age0,feh0,d0,AV0]).T
            else:
                p0 = np.array([m0,age0,feh0]).T
            if initial_burn:
                sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost,
                                                **kwargs)
                #ninitial = 300 #should this be parameter?
                pos, prob, state = sampler.run_mcmc(p0, ninitial) 
                wokinds = np.where((sampler.naccepted/ninitial > 0.15) &
                                   (sampler.naccepted/ninitial < 0.4))[0]
                i=1
                while len(wokinds)==0:
                    thresh = 0.15 - i*0.02
                    if thresh < 0:
                        raise RuntimeError('Initial burn has no acceptance?')
                    wokinds = np.where((sampler.naccepted/ninitial > thresh) &
                                       (sampler.naccepted/ninitial < 0.4))[0]
                    i += 1
                inds = rand.randint(len(wokinds),size=nwalkers)
                p0 = sampler.chain[wokinds[inds],:,:].mean(axis=1) #reset p0
                p0 *= (1 + rand.normal(size=p0.shape)*0.01)
        else:
            p0 = np.array(p0)
            p0 = rand.normal(size=(nwalkers,npars))*0.01 + p0.T[None,:]
            if self.fit_for_distance:
                p0[:,3] *= (1 + rand.normal(size=nwalkers)*0.5)
        
        sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost)
        pos, prob, state = sampler.run_mcmc(p0, nburn)
        sampler.reset()
        sampler.run_mcmc(pos, niter, rstate0=state)
        
        self._sampler = sampler
        return sampler

    def mag_plot(self, height=500, pix_width=20, spacing=20,
                 edge=0.1, figsize=(8,6)):
        
        bands = np.array(self.mag_errs.keys())
        weffs = np.array([WEFF[b] for b in bands])
        inds = np.argsort(weffs)
        bands = bands[inds]
        weffs = weffs[inds]

        q = 0.01
        minmag = min(np.min([self.samples['{}_mag'.format(b)].quantile(q) 
                             for b in bands]) - edge,
                     np.min([self.properties[b][0] - edge for b in bands]))
        maxmag = max(np.max([self.samples['{}_mag'.format(b)].quantile(1-q) 
                             for b in bands]) + edge,
                     np.max([self.properties[b][0] + edge for b in bands]))

        n_bands = len(bands)
        width = n_bands * (pix_width + spacing) + spacing
        mag_grid = np.linspace(minmag, maxmag, height)[::-1]

        image = np.zeros((height, width))

        plt.figure(figsize=figsize)

        mids = []
        for i,b in enumerate(bands):
            col1 = spacing*(i+1) + i*(pix_width)
            col2 = spacing*(i+1) + (i+1)*(pix_width)
            mids.append((col1 + col2)//2)
            vslice = image[:, col1:col2]

            kde = gaussian_kde(self.samples['{}_mag'.format(b)])
            pdf = kde(mag_grid)
            vslice += pdf[:, np.newaxis]

        extent = [0, image.shape[1], maxmag, minmag]
        plt.imshow(image, aspect='auto', cmap='binary', 
                   extent=extent, origin='lower')
        ax = plt.gca()
        ax.set_xticks(mids)
        ax.set_xticklabels(bands, fontsize=18);
        ax.set_ylabel('mag', fontsize=18)
        yticks = ax.get_yticks()
        ax.set_yticks(yticks[::-1])
        plt.tick_params(axis='y', labelsize=16)

        for i,(b,m) in enumerate(zip(bands,mids)):
            val, err = self.properties[b]
            plt.errorbar(m, val, err, marker='o', color='w', 
                         ms=4, lw=5, mec='w', mew=5)    
            plt.errorbar(m, val, err, marker='o', color='r', 
                         ms=4, lw=3, mec='r', mew=3)

        plt.title(self.name, fontsize=20)

        return plt.gcf()

    def triangle_plots(self, basename=None, format='png',
                       **kwargs):
        """Returns two triangle plots, one with physical params, one observational

        :param basename:
            If basename is provided, then plots will be saved as
            "[basename]_physical.[format]" and "[basename]_observed.[format]"

        :param format:
            Format in which to save figures (e.g., 'png' or 'pdf')

        :param **kwargs:
            Additional keyword arguments passed to :func:`StarModel.triangle`
            and :func:`StarModel.prop_triangle`

        :return:
             * Physical parameters triangle plot (mass, radius, Teff, feh, age, distance)
             * Observed properties triangle plot.
             
        """
        if self.fit_for_distance:
            fig1 = self.triangle(plot_datapoints=False,
                                 params=['mass','radius','Teff','logg','feh','age',
                                         'distance','AV'],
                                 **kwargs)
        else:
            fig1 = self.triangle(plot_datapoints=False,
                                 params=['mass','radius','Teff','feh','age'],
                                 **kwargs)
            

        if basename is not None:
            plt.savefig('{}_physical.{}'.format(basename,format))
            plt.close()
        fig2 = self.prop_triangle(**kwargs)
        if basename is not None:
            plt.savefig('{}_observed.{}'.format(basename,format))
            plt.close()
        return fig1, fig2

    def triangle(self, params=None, query=None, extent=0.999,
                 **kwargs):
        """
        Makes a nifty corner plot.

        Uses :func:`triangle.corner`.

        :param params: (optional)
            Names of columns (from :attr:`StarModel.samples`)
            to plot.  If ``None``, then it will plot samples
            of the parameters used in the MCMC fit-- that is,
            mass, age, [Fe/H], and optionally distance and A_V.

        :param query: (optional)
            Optional query on samples.

        :param extent: (optional)
            Will be appropriately passed to :func:`triangle.corner`.

        :param **kwargs:
            Additional keyword arguments passed to :func:`triangle.corner`.

        :return:
            Figure oject containing corner plot.
            
        """
        if triangle is None:
            raise ImportError('please run "pip install triangle_plot".')
        
        if params is None:
            if self.fit_for_distance:
                params = ['mass', 'age', 'feh', 'distance', 'AV']
            else:
                params = ['mass', 'age', 'feh']

        df = self.samples

        if query is not None:
            df = df.query(query)

        #convert extent to ranges, but making sure
        # that truths are in range.
        extents = []
        remove = []
        for i,par in enumerate(params):
            m = re.search('delta_(\w+)$',par)
            if m:
                if type(self) == BinaryStarModel:
                    b = m.group(1)
                    values = (df['{}_mag_B'.format(b)] - 
                              df['{}_mag_A'.format(b)])
                    df[par] = values
                else:
                    remove.append(i)
                    continue
                    
            else:
                values = df[par]
            qs = np.array([0.5 - 0.5*extent, 0.5 + 0.5*extent])
            minval, maxval = values.quantile(qs)
            if 'truths' in kwargs:
                datarange = maxval - minval
                if kwargs['truths'][i] < minval:
                    minval = kwargs['truths'][i] - 0.05*datarange
                if kwargs['truths'][i] > maxval:
                    maxval = kwargs['truths'][i] + 0.05*datarange
            extents.append((minval,maxval))
            
        [params.pop(i) for i in remove]

        fig = triangle.corner(df[params], labels=params, 
                               extents=extents, **kwargs)

        fig.suptitle(self.name, fontsize=22)
        return fig

    def prop_triangle(self, **kwargs):
        """
        Makes corner plot of only observable properties.

        The idea here is to compare the predictions of the samples
        with the actual observed data---this can be a quick way to check
        if there are outlier properties that aren't predicted well
        by the model.

        :param **kwargs:
            Keyword arguments passed to :func:`StarModel.triangle`.

        :return:
            Figure object containing corner plot.
         
        """
        truths = []
        params = []
        for p in self.properties:
            try:
                val, err = self.properties[p]
            except:
                continue

            if p in self.ic.bands:
                params.append('{}_mag'.format(p))
                truths.append(val)
            elif p=='parallax':
                params.append('distance')
                truths.append(1/(val/1000.))
            else:
                params.append(p)
                truths.append(val)
        return self.triangle(params, truths=truths, **kwargs)
        

    @property
    def sampler(self):
        """
        Sampler object from MCMC run.
        """
        if hasattr(self,'_sampler'):
            return self._sampler
        else:
            raise AttributeError('MCMC must be run to access sampler')

    def _make_samples(self):

        if not self.use_emcee:
            chain = np.loadtxt('{}post_equal_weights.dat'.format(self._mnest_basename))

            #for purposes of unit test, sometimes there will be 1-length chain...
            if chain.ndim==1:
                chain = np.array([chain])

            mass = chain[:,0]
            age = chain[:,1]
            feh = chain[:,2]
            distance = chain[:,3]
            AV = chain[:,4]
            lnprob = chain[:,-1]

        else:
            #select out only walkers with > 0.15 acceptance fraction
            ok_walkers = self.sampler.acceptance_fraction > 0.15

            mass = self.sampler.chain[ok_walkers, :, 0].ravel()
            age = self.sampler.chain[ok_walkers, :, 1].ravel()
            feh = self.sampler.chain[ok_walkers, :, 2].ravel()

            if self.fit_for_distance:
                distance = self.sampler.chain[ok_walkers, :, 3].ravel()
                AV = self.sampler.chain[ok_walkers, :, 4].ravel()
            else:
                distance = None
                AV = 0

            lnprob = self.sampler.lnprobability[ok_walkers, :].ravel()

        df = self.ic(mass, age, feh, 
                     distance=distance, AV=AV)
        df['age'] = age
        df['feh'] = feh

        if self.fit_for_distance:
            df['distance'] = distance
            df['AV'] = AV

        df['lnprob'] = lnprob

        self._samples = df.copy()
        
    @property
    def samples(self):
        """Dataframe with samples drawn from isochrone according to posterior

        Columns include both the sampling parameters from the MCMC
        fit (mass, age, Fe/H, [distance, A_V]), and also evaluation
        of the :class:`Isochrone` at each of these sample points---this
        is how chains of physical/observable parameters get produced.
        
        """
        if not hasattr(self,'sampler') and self._samples is None:
            raise AttributeError('Must run MCMC (or load from file) '+
                                 'before accessing samples')
        
        if self._samples is not None:
            df = self._samples
        else:
            self._make_samples()
            df = self._samples

        return df

    def random_samples(self, n):
        """
        Returns a random sampling of given size from the existing samples.

        :param n:
            Number of samples

        :return:
            :class:`pandas.DataFrame` of length ``n`` with random samples.
        """
        samples = self.samples
        inds = rand.randint(len(samples),size=int(n))

        newsamples = samples.iloc[inds]
        newsamples.reset_index(inplace=True)
        return newsamples


    def prop_samples(self,prop,return_values=True,conf=0.683):
        """Returns samples of given property, based on MCMC sampling

        :param prop:
            Name of desired property.  Must be column of ``self.samples``.

        :param return_values: (optional)
            If ``True`` (default), then also return (median, lo_err, hi_err)
            corresponding to desired credible interval.

        :param conf: (optional)
            Desired quantile for credible interval.  Default = 0.683.

        :return:
            :class:`np.ndarray` of desired samples

        :return: 
            Optionally also return summary statistics (median, lo_err, hi_err),
            if ``returns_values == True`` (this is default behavior)
        
        """
        samples = self.samples[prop].values
        
        if return_values:
            sorted = np.sort(samples)
            med = np.median(samples)
            n = len(samples)
            lo_ind = int(n*(0.5 - conf/2))
            hi_ind = int(n*(0.5 + conf/2))
            lo = med - sorted[lo_ind]
            hi = sorted[hi_ind] - med
            return samples, (med,lo,hi)
        else:
            return samples

    def plot_samples(self,prop,fig=None,label=True,
                     histtype='step',bins=50,lw=3,
                     **kwargs):
        """Plots histogram of samples of desired property.

        :param prop:
            Desired property (must be legit column of samples)
            
        :param fig:
              Argument for :func:`plotutils.setfig` (``None`` or int).

        :param histtype, bins, lw:
             Passed to :func:`plt.hist`.

        :param **kwargs:
            Additional keyword arguments passed to `plt.hist`

        :return:
            Figure object.
        """
        setfig(fig)
        samples,stats = self.prop_samples(prop)
        fig = plt.hist(samples,bins=bins,normed=True,
                 histtype=histtype,lw=lw,**kwargs)
        plt.xlabel(prop)
        plt.ylabel('Normalized count')
        
        if label:
            med,lo,hi = stats
            plt.annotate('$%.2f^{+%.2f}_{-%.2f}$' % (med,hi,lo),
                         xy=(0.7,0.8),xycoords='axes fraction',fontsize=20)

        return fig
            
    def save_hdf(self, filename, path='', overwrite=False, append=False):
        """Saves object data to HDF file (only works if MCMC is run)

        Samples are saved to /samples location under given path,
        and object properties are also attached, so suitable for
        re-loading via :func:`StarModel.load_hdf`.
        
        :param filename:
            Name of file to save to.  Should be .h5 file.

        :param path: (optional)
            Path within HDF file structure to save to.

        :param overwrite: (optional)
            If ``True``, delete any existing file by the same name
            before writing.

        :param append: (optional)
            If ``True``, then if a file exists, then just the path
            within the file will be updated.
        """
        
        if os.path.exists(filename):
            store = pd.HDFStore(filename)
            if path in store:
                store.close()
                if overwrite:
                    os.remove(filename)
                elif not append:
                    raise IOError('{} in {} exists.  Set either overwrite or append option.'.format(path,filename))
            else:
                store.close()

        self.samples.to_hdf(filename, '{}/samples'.format(path))

        store = pd.HDFStore(filename)
        attrs = store.get_storer('{}/samples'.format(path)).attrs
        attrs.properties = self.properties
        attrs.ic_type = type(self.ic)
        attrs.maxAV = self.maxAV
        attrs.max_distance = self.max_distance
        attrs.min_logg = self.min_logg

        attrs.use_emcee = self.use_emcee
        attrs._mnest_basename = self._mnest_basename

        attrs.name = self.name

        store.close()

    @classmethod
    def load_hdf(cls, filename, path='', name=None):
        """
        A class method to load a saved StarModel from an HDF5 file.

        File must have been created by a call to :func:`StarModel.save_hdf`.

        :param filename:
            H5 file to load.

        :param path: (optional)
            Path within HDF file.

        :return:
            :class:`StarModel` object.
        """
        store = pd.HDFStore(filename)
        try:
            samples = store['{}/samples'.format(path)]
            attrs = store.get_storer('{}/samples'.format(path)).attrs        
        except:
            store.close()
            raise
        properties = attrs.properties
        maxAV = attrs.maxAV
        max_distance = attrs.max_distance
        min_logg = attrs.min_logg
        ic_type = attrs.ic_type
        use_emcee = attrs.use_emcee
        basename = attrs._mnest_basename
        
        if name is None:
            try:
                name = attrs.name
            except:
                name = ''

        store.close()

        #ic = ic_type() don't need to initialize anymore

        mod = cls(ic_type, maxAV=maxAV, max_distance=max_distance,
                  use_emcee=use_emcee, name=name,
                  **properties)
        mod._samples = samples
        mod._mnest_basename = basename
        return mod


class BinaryStarModel(StarModel):
    """
    Object used to fit two stars at the same distance to given observed properties

    Initialize the same way as :class:`StarModel`. 

    Difference between this object and a regular :class:`StarModel` is that
    the fit parameters include two masses: ``mass_A`` and ``mass_B`` instead
    of just one.

    Notably, this object can also take additional 
    ``delta_mag`` properties, representing contrast measurements of a companion
    star.  These should be called, e.g., ``delta_r``.  This will be an additional
    constraint in the model fitting.  All the other provided apparent magnitudes
    should be the total combined light of the two stars.




    """
    def __init__(self, *args, **kwargs):
        super(BinaryStarModel, self).__init__(*args, **kwargs)

        self.n_params = 6

    def lnlike(self, p):
        """Log-likelihood of model at given parameters

        
        :param p: 
            mass_A, mass_B, log10(age), feh, [distance, A_V (extinction)].
            Final two should only be provided if ``self.fit_for_distance``
            is ``True``; that is, apparent magnitudes are provided.
            
        :return:
           log-likelihood.  Will be -np.inf if values out of range.
        
        """
        if not self._props_cleaned:
            self._clean_props()

        if not self.use_emcee:
            fit_for_distance = True
            mass_A, mass_B, age, feh, dist, AV = (p[0], p[1], p[2],
                                                  p[3], p[4], p[5])
        else:
            if len(p)==6:
                fit_for_distance = True
                mass_A, mass_B, age, feh, dist, AV = p
            elif len(p)==4:
                fit_for_distance = False
                mass_A, mass_B, age, feh = p
                        
        #keep values in range; enforce mass_A > mass_B
        if mass_A < self.ic.minmass or mass_A > self.ic.maxmass \
           or mass_B < self.ic.minmass or mass_B > self.ic.maxmass \
           or mass_B > mass_A \
           or age < self.ic.minage or age > self.ic.maxage \
           or feh < self.ic.minfeh or feh > self.ic.maxfeh:
            return -np.inf
        if fit_for_distance:
            if dist < 0 or AV < 0 or dist > self.max_distance:
                return -np.inf
            if AV > self.maxAV:
                return -np.inf

        if self.min_logg is not None:
            logg = self.ic.logg(mass_A,age,feh)
            if logg < self.min_logg:
                return -np.inf        

        logl = 0
        for prop in self.properties.keys():
            try:
                val,err = self.properties[prop]
            except TypeError:
                #property not appropriate for fitting (e.g. no error provided)
                continue
            m = re.search('delta_(\w+)',prop)
            if prop in self.ic.bands:
                if not fit_for_distance:
                    raise ValueError('must fit for mass, age, feh, dist, A_V '+
                                     'if apparent magnitudes provided.')
                mod_A = self.ic.mag[prop](mass_A,age,feh) + 5*np.log10(dist) - 5
                mod_B = self.ic.mag[prop](mass_B,age,feh) + 5*np.log10(dist) - 5
                A = AV*EXTINCTION[prop]
                mod_A += A
                mod_B += A
                mod = addmags(mod_A, mod_B)
            elif m:
                band = m.group(1)
                mod_A = self.ic.mag[band](mass_A,age,feh) + 5*np.log10(dist) - 5
                mod_B = self.ic.mag[band](mass_B,age,feh) + 5*np.log10(dist) - 5
                A = AV*EXTINCTION[band]
                mod_A += A
                mod_B += A
                mod = mod_B - mod_A
            elif prop=='feh':
                mod = feh
            elif prop=='mass_B':
                mod = mass_B
            elif prop=='parallax':
                mod = 1./dist * 1000
            else:
                mod = getattr(self.ic,prop)(mass_A,age,feh)

            logl += -(val-mod)**2/(2*err**2) + np.log(1/(err*np.sqrt(2*np.pi)))


        if np.isnan(logl):
            logl = -np.inf


        return logl
        
    def lnprior(self, mass_A, mass_B, age, feh, 
                distance=None, AV=None, use_local_fehprior=True):
        lnpr = super(BinaryStarModel,self).lnprior(mass_A, age, feh, distance, AV,
                                                  use_local_fehprior=use_local_fehprior)
        q = mass_B / mass_A
        lnpr += np.log(q_prior(q, mass_A))
        return lnpr

    def lnpost(self, p, use_local_fehprior=True):
        if not self.use_emcee:
            mass_A,mass_B,age,feh,dist,AV = (p[0], p[1], p[2],
                                             p[3], p[4], p[5])
        else:
            if len(p)==6:
                fit_for_distance = True
                mass_A,mass_B,age,feh,dist,AV = p
            elif len(p)==4:
                fit_for_distance = False
                mass_A,mass_B,age,feh = p
                dist = None
                AV = None

        return (self.lnlike(p) + 
                self.lnprior(mass_A, mass_B, age, feh, dist, AV,
                             use_local_fehprior=use_local_fehprior))
        

    def maxlike(self,nseeds=50):
        """Returns the best-fit parameters, choosing the best of multiple starting guesses

        :param nseeds: (optional)
            Number of starting guesses, uniformly distributed throughout
            allowed ranges.  Default=50.

        :return:
            list of best-fit parameters: ``[mA,mB,age,feh,[distance,A_V]]``.
            Note that distance and A_V values will be meaningless unless
            magnitudes are present in ``self.properties``.
        
        """
        mA_0,age0,feh0 = self.ic.random_points(nseeds)
        mB_0,foo1,foo2 = self.ic.random_points(nseeds)
        mA_fixed = np.maximum(mA_0,mB_0)
        mB_fixed = np.minimum(mA_0,mB_0)
        mA_0, mB_0 = (mA_fixed, mB_fixed)

        d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nseeds))
        AV0 = rand.uniform(0,self.maxAV,size=nseeds)

        

        costs = np.zeros(nseeds)

        if self.fit_for_distance:
            pfits = np.zeros((nseeds,6))
        else:
            pfits = np.zeros((nseeds,4))
            
        def fn(p): #fmin is a function *minimizer*
            return -1*self.lnpost(p)
        
        for i,mA,mB,age,feh,d,AV in zip(range(nseeds),
                                    mA_0,mB_0,age0,feh0,d0,AV0):
                if self.fit_for_distance:
                    pfit = scipy.optimize.fmin(fn,[mA,mB,age,feh,d,AV],disp=False)
                else:
                    pfit = scipy.optimize.fmin(fn,[mA,mB,age,feh],disp=False)
                pfits[i,:] = pfit
                costs[i] = self.lnpost(pfit)

        return pfits[np.argmax(costs),:]

    def mnest_prior(self, cube, ndim, nparams):
        cube[0] = (self.ic.maxmass - self.ic.minmass)*cube[0] + self.ic.minmass
        cube[1] = (cube[0] - self.ic.minmass)*cube[1] + self.ic.minmass
        cube[2] = (self.ic.maxage - self.ic.minage)*cube[2] + self.ic.minage
        cube[3] = (self.ic.maxfeh - self.ic.minfeh)*cube[3] + self.ic.minfeh
        cube[4] = cube[4]*self.max_distance
        cube[5] = cube[5]*self.maxAV
        
    def fit_multinest(self, basename='chains/binary-', **kwargs):
        super(BinaryStarModel, self).fit_multinest(basename=basename, **kwargs)

    def fit_mcmc(self,nwalkers=200,nburn=100,niter=200,
                 p0=None,initial_burn=None,
                 ninitial=100, loglike_kwargs=None,
                 **kwargs):
        """Fits stellar model using MCMC.

        See :func:`StarModel.fit_mcmc`
        """

        #clear any saved _samples
        if self._samples is not None:
            self._samples = None
            

        if self.fit_for_distance:
            npars = 6
            if initial_burn is None:
                initial_burn = True
        else:
            if initial_burn is None:
                initial_burn = False
            npars = 4

        if p0 is None:
            mA_0,age0,feh0 = self.ic.random_points(nwalkers)
            mB_0,foo1,foo2 = self.ic.random_points(nwalkers)
            mA_fixed = np.maximum(mA_0,mB_0)
            mB_fixed = np.minimum(mA_0,mB_0)
            mA_0, mB_0 = (mA_fixed, mB_fixed)

            d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nwalkers))
            AV0 = rand.uniform(0,self.maxAV,size=nwalkers)
            if self.fit_for_distance:
                p0 = np.array([mA_0,mB_0,age0,feh0,d0,AV0]).T
            else:
                p0 = np.array([mA_0,mB_0,age0,feh0]).T
            if initial_burn:
                sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost,
                                                **kwargs)
                #ninitial = 300 #should this be parameter?
                pos, prob, state = sampler.run_mcmc(p0, ninitial) 
                wokinds = np.where((sampler.naccepted/ninitial > 0.15) &
                                   (sampler.naccepted/ninitial < 0.4))[0]
                i=1
                while len(wokinds)==0:
                    thresh = 0.15 - i*0.02
                    if thresh < 0:
                        raise RuntimeError('Initial burn has no acceptance?')
                    wokinds = np.where((sampler.naccepted/ninitial > thresh) &
                                       (sampler.naccepted/ninitial < 0.4))[0]
                    i += 1
                inds = rand.randint(len(wokinds),size=nwalkers)
                p0 = sampler.chain[wokinds[inds],:,:].mean(axis=1) #reset p0
                p0 *= (1 + rand.normal(size=p0.shape)*0.01)
        else:
            p0 = np.array(p0)
            p0 = rand.normal(size=(nwalkers,npars))*0.01 + p0.T[None,:]
            if self.fit_for_distance:
                p0[:,4] *= (1 + rand.normal(size=nwalkers)*0.5)
        
        sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost)
        pos, prob, state = sampler.run_mcmc(p0, nburn)
        sampler.reset()
        sampler.run_mcmc(pos, niter, rstate0=state)
        
        self._sampler = sampler
        return sampler

    def triangle_plots(self, basename=None, format='png',
                       **kwargs):
        """Returns two triangle plots, one with physical params, one observational

        :param basename:
            If basename is provided, then plots will be saved as
            "[basename]_physical.[format]" and "[basename]_observed.[format]"

        :param format:
            Format in which to save figures (e.g., 'png' or 'pdf')

        :param **kwargs:
            Additional keyword arguments passed to :func:`StarModel.triangle`
            and :func:`StarModel.prop_triangle`

        :return:
             * Physical parameters triangle plot (mass_A, mass_B, radius, Teff, feh, age, distance)
             * Observed properties triangle plot.
             
        """
        fig1 = self.triangle(plot_datapoints=False,
                            params=['mass_A', 'mass_B','radius','Teff','logg','feh','age',
                                    'distance', 'AV'],
                            **kwargs)
        if basename is not None:
            plt.savefig('{}_physical.{}'.format(basename,format))
            plt.close()
        fig2 = self.prop_triangle(**kwargs)
        if basename is not None:
            plt.savefig('{}_observed.{}'.format(basename,format))
            plt.close()
        return fig1, fig2

    def triangle(self, params=None, **kwargs):
        """
        Makes a nifty corner plot.

        Uses :func:`triangle.corner`.

        :param params: (optional)
            Names of columns (from :attr:`StarModel.samples`)
            to plot.  If ``None``, then it will plot samples
            of the parameters used in the MCMC fit-- that is,
            mass, age, [Fe/H], and optionally distance and A_V.

        :param query: (optional)
            Optional query on samples.

        :param extent: (optional)
            Will be appropriately passed to :func:`triangle.corner`.

        :param **kwargs:
            Additional keyword arguments passed to :func:`triangle.corner`.

        :return:
            Figure oject containing corner plot.
            
        """
        if params is None:
            params = ['mass_A', 'mass_B', 'age', 'feh', 'distance', 'AV']

        super(BinaryStarModel, self).triangle(params=params, **kwargs)


    def _make_samples(self):

        if not self.use_emcee:
            chain = np.loadtxt('{}post_equal_weights.dat'.format(self._mnest_basename))

            #for purposes of unit test, sometimes there will be 1-length chain...
            if chain.ndim==1:
                chain = np.array([chain])

            mass_A = chain[:,0]
            mass_B = chain[:,1]
            age = chain[:,2]
            feh = chain[:,3]
            distance = chain[:,4]
            AV = chain[:,5]
            lnprob = chain[:,-1]

        else:
            #select out legit walkers
            ok_walkers = self.sampler.acceptance_fraction > 0.15

            mass_A = self.sampler.chain[ok_walkers,:,0].ravel()
            mass_B = self.sampler.chain[ok_walkers,:,1].ravel()
            age = self.sampler.chain[ok_walkers,:,2].ravel()
            feh = self.sampler.chain[ok_walkers,:,3].ravel()

            if self.fit_for_distance:
                distance = self.sampler.chain[ok_walkers,:,4].ravel()
                AV = self.sampler.chain[ok_walkers,:,5].ravel()
            else:
                distance = None
                AV = 0
            lnprob = self.sampler.lnprobability[ok_walkers,:].ravel()


        df = self.ic(mass_A, age, feh, 
                       distance=distance, AV=AV)
        df_B = self.ic(mass_B, age, feh, 
                       distance=distance, AV=AV)

        for col in df_B.columns:
            m = re.search('_mag$', col)
            if m:
                df['{}_A'.format(col)] = df[col]
                df['{}_B'.format(col)] = df_B[col]
                df[col] = addmags(df[col], df_B[col])


        df['mass_A'] = df['mass']
        df.drop('mass', axis=1, inplace=True)
        df['mass_B'] = df_B['mass']
        df['radius_B'] = df_B['radius']
        df['Teff_B'] = df_B['Teff']
        df['logg_B'] = df_B['logg']
        df['logL_B'] = df_B['logL']
        df['age'] = age
        df['feh'] = feh

        if self.fit_for_distance:
            df['distance'] = distance
            df['AV'] = AV

        df['lnprob'] = lnprob

        self._samples = df.copy()



class TripleStarModel(StarModel):
    """Just like BinaryStarModel but for three.

    Parameters now include mass_A, mass_B, and mass_C
    """
    def __init__(self, *args, **kwargs):
        super(TripleStarModel, self).__init__(*args, **kwargs)

        self.n_params = 7

    def lnlike(self, p):
        """Log-likelihood of model at given parameters

        
        :param p: 
            mass_A, mass_B, mass_C, log10(age), feh, [distance, A_V (extinction)].
            Final two should only be provided if ``self.fit_for_distance``
            is ``True``; that is, apparent magnitudes are provided.
            
        :return:
           log-likelihood.  Will be -np.inf if values out of range.
        
        """
        if not self._props_cleaned:
            self._clean_props()

        if not self.use_emcee:
            fit_for_distance = True
            mass_A, mass_B, mass_C, age, feh, dist, AV = (p[0], p[1], p[2],
                                                          p[3], p[4], p[5],
                                                          p[6])
        else:
            if len(p)==7:
                fit_for_distance = True
                mass_A, mass_B, mass_C, age, feh, dist, AV = p
            elif len(p)==5:
                fit_for_distance = False
                mass_A, mass_B, mass_C, age, feh = p
                        
        #keep values in range; enforce mass_A > mass_B > mass_C
        if mass_A < self.ic.minmass or mass_A > self.ic.maxmass \
           or mass_B < self.ic.minmass or mass_B > self.ic.maxmass \
           or mass_C < self.ic.minmass or mass_C > self.ic.maxmass \
           or mass_B > mass_A \
           or mass_C > mass_B or mass_C > mass_A \
           or age < self.ic.minage or age > self.ic.maxage \
           or feh < self.ic.minfeh or feh > self.ic.maxfeh:
            return -np.inf
        if fit_for_distance:
            if dist < 0 or AV < 0 or dist > self.max_distance:
                return -np.inf
            if AV > self.maxAV:
                return -np.inf

        if self.min_logg is not None:
            logg = self.ic.logg(mass_A,age,feh)
            if logg < self.min_logg:
                return -np.inf

        logl = 0
        for prop in self.properties.keys():
            try:
                val,err = self.properties[prop]
            except TypeError:
                #property not appropriate for fitting (e.g. no error provided)
                continue
            if prop in self.ic.bands:
                if not fit_for_distance:
                    raise ValueError('must fit for mass_A, mass_B, mass_C, age, feh, dist,'+ 
                                     'A_V if apparent magnitudes provided.')
                mods = self.ic.mag[prop]([mass_A, mass_B, mass_C], 
                                         age, feh) + 5*np.log10(dist) - 5
                A = AV*EXTINCTION[prop]
                mods += A
                mod = addmags(*mods)
            elif re.search('delta_',prop):
                continue
            elif prop=='feh':
                mod = feh
            elif prop=='mass_B':
                mod = mass_B
            elif prop=='mass_C':
                mod = mass_C
            elif prop=='parallax':
                mod = 1./dist * 1000
            else:
                mod = getattr(self.ic,prop)(mass_A,age,feh)

            logl += -(val-mod)**2/(2*err**2) + np.log(1/(err*np.sqrt(2*np.pi)))



        if np.isnan(logl):
            logl = -np.inf

        return logl

    def lnprior(self, mass_A, mass_B, mass_C, age, feh, 
                distance=None, AV=None, use_local_fehprior=True):
        lnpr = super(TripleStarModel,self).lnprior(mass_A, age, feh, distance, AV,
                                                  use_local_fehprior=use_local_fehprior)
        q1 = mass_B / mass_A
        q2 = mass_C / mass_B
        lnpr += np.log(q_prior(q1, mass_A))
        lnpr += np.log(q_prior(q2, mass_B))
        return lnpr


    def lnpost(self, p, use_local_fehprior=True):
        if not self.use_emcee:
            mass_A,mass_B,mass_C,age,feh,dist,AV = (p[0], p[1], p[2],
                                                    p[3], p[4], p[5], 
                                                    p[6])
        else:
            if len(p)==7:
                fit_for_distance = True
                mass_A,mass_B,mass_C,age,feh,dist,AV = p
            elif len(p)==5:
                fit_for_distance = False
                mass_A,mass_B,mass_C,age,feh = p
                dist = None
                AV = None

        return (self.lnlike(p) + 
                self.lnprior(mass_A, mass_B, mass_C, age, feh, dist, AV,
                             use_local_fehprior=use_local_fehprior))
        
    def maxlike(self,nseeds=50):
        """Returns the best-fit parameters, choosing the best of multiple starting guesses

        :param nseeds: (optional)
            Number of starting guesses, uniformly distributed throughout
            allowed ranges.  Default=50.

        :return:
            list of best-fit parameters: ``[mA,mB,age,feh,[distance,A_V]]``.
            Note that distance and A_V values will be meaningless unless
            magnitudes are present in ``self.properties``.
        
        """
        mA_0,age0,feh0 = self.ic.random_points(nseeds)
        mB_0,foo1,foo2 = self.ic.random_points(nseeds)
        mC_0,foo3,foo4 = self.ic.random_points(nseeds)
        m_all = np.sort(np.array([mA_0, mB_0, mC_0]), axis=0)
        mA_0, mB_0, mC_0 = (m_all[0,:], m_all[1,:], m_all[2,:])

        d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nseeds))
        AV0 = rand.uniform(0,self.maxAV,size=nseeds)

        

        costs = np.zeros(nseeds)

        if self.fit_for_distance:
            pfits = np.zeros((nseeds,7))
        else:
            pfits = np.zeros((nseeds,5))
            
        def fn(p): #fmin is a function *minimizer*
            return -1*self.lnpost(p)
        
        for i,mA,mB,mC,age,feh,d,AV in zip(range(nseeds),
                                    mA_0,mB_0,mC_0,age0,feh0,d0,AV0):
                if self.fit_for_distance:
                    pfit = scipy.optimize.fmin(fn,[mA,mB,mC,age,feh,d,AV],disp=False)
                else:
                    pfit = scipy.optimize.fmin(fn,[mA,mB,mC,age,feh],disp=False)
                pfits[i,:] = pfit
                costs[i] = self.lnpost(pfit)

        return pfits[np.argmax(costs),:]

            
    def mnest_prior(self, cube, ndim, nparams):
        cube[0] = (self.ic.maxmass - self.ic.minmass)*cube[0] + self.ic.minmass
        cube[1] = (cube[0] - self.ic.minmass)*cube[1] + self.ic.minmass
        cube[2] = (cube[1] - self.ic.minmass)*cube[2] + self.ic.minmass
        cube[3] = (self.ic.maxage - self.ic.minage)*cube[3] + self.ic.minage
        cube[4] = (self.ic.maxfeh - self.ic.minfeh)*cube[4] + self.ic.minfeh
        cube[5] = cube[5]*self.max_distance
        cube[6] = cube[6]*self.maxAV
        
    def fit_multinest(self, basename='chains/triple-', **kwargs):
        super(TripleStarModel, self).fit_multinest(basename=basename, **kwargs)    

    def fit_mcmc(self,nwalkers=200,nburn=100,niter=200,
                 p0=None,initial_burn=None,
                 ninitial=100, loglike_kwargs=None,
                 **kwargs):
        """Fits stellar model using MCMC.

        See :func:`StarModel.fit_mcmc`.
            
        """

        #clear any saved _samples
        if self._samples is not None:
            self._samples = None
            

        if self.fit_for_distance:
            npars = 7
            if initial_burn is None:
                initial_burn = True
        else:
            if initial_burn is None:
                initial_burn = False
            npars = 5

        if p0 is None:
            mA_0,age0,feh0 = self.ic.random_points(nwalkers)
            mB_0,foo1,foo2 = self.ic.random_points(nwalkers)
            mC_0,foo3,foo4 = self.ic.random_points(nwalkers)
            m_all = np.sort(np.array([mA_0, mB_0, mC_0]), axis=0)
            mA_0, mB_0, mC_0 = (m_all[0,:], m_all[1,:], m_all[2,:])

            d0 = 10**(rand.uniform(0,np.log10(self.max_distance),size=nwalkers))
            AV0 = rand.uniform(0,self.maxAV,size=nwalkers)
            if self.fit_for_distance:
                p0 = np.array([mA_0,mB_0,mC_0,age0,feh0,d0,AV0]).T
            else:
                p0 = np.array([mA_0,mB_0,mC_0,age0,feh0]).T
            if initial_burn:
                sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost,
                                                **kwargs)
                #ninitial = 300 #should this be parameter?
                pos, prob, state = sampler.run_mcmc(p0, ninitial) 
                wokinds = np.where((sampler.naccepted/ninitial > 0.15) &
                                   (sampler.naccepted/ninitial < 0.4))[0]
                i=1
                while len(wokinds)==0:
                    thresh = 0.15 - i*0.02
                    if thresh < 0:
                        raise RuntimeError('Initial burn has no acceptance?')
                    wokinds = np.where((sampler.naccepted/ninitial > thresh) &
                                       (sampler.naccepted/ninitial < 0.4))[0]
                    i += 1
                inds = rand.randint(len(wokinds),size=nwalkers)
                p0 = sampler.chain[wokinds[inds],:,:].mean(axis=1) #reset p0
                p0 *= (1 + rand.normal(size=p0.shape)*0.01)
        else:
            p0 = np.array(p0)
            p0 = rand.normal(size=(nwalkers,npars))*0.01 + p0.T[None,:]
            if self.fit_for_distance:
                p0[:,5] *= (1 + rand.normal(size=nwalkers)*0.5) #distance
        
        sampler = emcee.EnsembleSampler(nwalkers,npars,self.lnpost)
        pos, prob, state = sampler.run_mcmc(p0, nburn)
        sampler.reset()
        sampler.run_mcmc(pos, niter, rstate0=state)
        
        self._sampler = sampler
        return sampler

    def triangle_plots(self, basename=None, format='png',
                       **kwargs):
        """Returns two triangle plots, one with physical params, one observational

        :return:
             * Physical parameters triangle plot (mass_A, mass_B, mass_C, radius, 
                Teff, feh, age, distance)
             * Observed properties triangle plot.
             
        """
        fig1 = self.triangle(plot_datapoints=False,
                            params=['mass_A', 'mass_B', 'mass_C', 'radius',
                                    'Teff','logg','feh','age','distance','AV'],
                            **kwargs)
        if basename is not None:
            plt.savefig('{}_physical.{}'.format(basename,format))
            plt.close()
        fig2 = self.prop_triangle(**kwargs)
        if basename is not None:
            plt.savefig('{}_observed.{}'.format(basename,format))
            plt.close()
        return fig1, fig2

    def triangle(self, params=None, **kwargs):
        """
        Makes a nifty corner plot.

        """
        if params is None:
            params = ['mass_A', 'mass_B', 'mass_C', 
                      'age', 'feh', 'distance', 'AV']

        super(TripleStarModel, self).triangle(params=params, **kwargs)


    def _make_samples(self):
        
        if not self.use_emcee:
            chain = np.loadtxt('{}post_equal_weights.dat'.format(self._mnest_basename))

            #for purposes of unit test, sometimes there will be 1-length chain...
            if chain.ndim==1:
                chain = np.array([chain])

            mass_A = chain[:,0]
            mass_B = chain[:,1]
            mass_C = chain[:,2]
            age = chain[:,3]
            feh = chain[:,4]
            distance = chain[:,5]
            AV = chain[:,6]
            lnprob = chain[:,-1]

        else:
            ok_walkers = self.sampler.acceptance_fraction > 0.15

            mass_A = self.sampler.chain[ok_walkers,:,0].ravel()
            mass_B = self.sampler.chain[ok_walkers,:,1].ravel()
            mass_C = self.sampler.chain[ok_walkers,:,2].ravel()
            age = self.sampler.chain[ok_walkers,:,3].ravel()
            feh = self.sampler.chain[ok_walkers,:,4].ravel()

            if self.fit_for_distance:
                distance = self.sampler.chain[ok_walkers,:,5].ravel()
                AV = self.sampler.chain[ok_walkers,:,6].ravel()
            else:
                distance = None
                AV = 0
            
            lnprob = self.sampler.lnprobability[ok_walkers,:].ravel()

        df = self.ic(mass_A, age, feh, 
                       distance=distance, AV=AV)
        df_B = self.ic(mass_B, age, feh, 
                       distance=distance, AV=AV)
        df_C = self.ic(mass_C, age, feh, 
                       distance=distance, AV=AV)

        for col in df_B.columns:
            m = re.search('_mag$', col)
            if m:
                df['{}_A'.format(col)] = df[col]
                df['{}_B'.format(col)] = df_B[col]
                df['{}_C'.format(col)] = df_C[col]
                df[col] = addmags(df[col], df_B[col], df_C[col])

        df['mass_A'] = df['mass']
        df.drop('mass', axis=1, inplace=True)
        df['mass_B'] = df_B['mass']
        df['mass_C'] = df_C['mass']
        df['radius_B'] = df_B['radius']
        df['radius_C'] = df_C['radius']
        df['Teff_B'] = df_B['Teff']
        df['Teff_C'] = df_C['Teff']
        df['logg_B'] = df_B['logg']
        df['logg_C'] = df_C['logg']
        df['logL_B'] = df_B['logL']
        df['logL_C'] = df_C['logL']
        df['age'] = age
        df['feh'] = feh

        if self.fit_for_distance:
            df['distance'] = distance
            df['AV'] = AV

        df['lnprob'] = lnprob

        self._samples = df.copy()

#class MultipleStarModel(StarModel, BinaryStarModel, TripleStarModel):
#    """
#    StarModel where N_stars (1,2, or 3) is a parameter to estimate.
#    """

#    def binary_loglike(self, *args, **kwargs):
#        return BinaryStarModel.loglike(self, *args, **kwargs)

#    def triple_loglike(self, *args, **kwargs):
#        return TripleStarModel.loglike(self, *args, **kwargs)
        

#### Utility functions #####


def addmags(*mags):
    tot=0
    for mag in mags:
        tot += 10**(-0.4*mag)
    return -2.5*np.log10(tot)
    
def q_prior(q, m=1, gamma=0.3, qmin=0.1):
    """Default prior on mass ratio q ~ q^gamma
    """
    if q < qmin or q > 1:
        return 0
    C = 1/(1/(gamma+1)*(1 - qmin**(gamma+1)))
    return C*q**gamma

def salpeter_prior(m,alpha=-2.35,minmass=0.1,maxmass=10):
    C = (1+alpha)/(maxmass**(1+alpha)-minmass**(1+alpha))
    if m < minmass or m > maxmass:
        return 0
    else:
        return C*m**(alpha)

def local_fehdist(feh):
    """feh PDF based on local SDSS distribution
    
    From Jo Bovy:
    https://github.com/jobovy/apogee/blob/master/apogee/util/__init__.py#L3
    2D gaussian fit based on Casagrande (2011)
    """
    fehdist= 0.8/0.15*np.exp(-0.5*(feh-0.016)**2./0.15**2.)\
        +0.2/0.22*np.exp(-0.5*(feh+0.15)**2./0.22**2.)

    return fehdist



import os, re, sys
import pandas as pd
import numpy as np

from scipy.interpolate import LinearNDInterpolator as interpnd
import numpy.random as rand
import matplotlib.pyplot as plt

try:
    from astropy import constants as const

    #Define useful constants
    G = const.G.cgs.value
    MSUN = const.M_sun.cgs.value
    RSUN = const.R_sun.cgs.value
except ImportError:
    G = 6.67e-11
    MSUN = 1.99e33
    RSUN = 6.96e10


from .extinction import EXTINCTION, LAMBDA_EFF, extcurve, extcurve_0
#from ..isochrone import Isochrone
from .config import ISOCHRONES
from .interp import interp_value, interp_values
from .grid import ModelGrid


class Isochrone(object):
    """
    Basic isochrone class. Everything is a function of mass, log(age), Fe/H.

    Can be instantiated directly, but will typically be used with a pre-defined
    subclass, such as :class:`dartmouth.Dartmouth_Isochrone`.  All parameters
    must be array-like objects of the same length, with the exception of ``mags``,
    which is a dictionary of such array-like objects.

    :param m_ini:
        Array of initial mass values [msun].
    :type m_ini: array-like

    :param age: 
        log10(age) [yr]

    :param feh:
        Metallicity [dex]

    :param m_act: 
        Actual mass; same as m_ini if mass loss not implemented [msun]

    :param logL:
        log10(luminosity) [solar units]

    :param Teff:
        Effective temperature [K]

    :param logg:
        log10(surface gravity) [cgs]

    :param mags: 
        Dictionary of absolute magnitudes in different bands
    :type mags: ``dict``
        
    :param tri:
        Triangulation object used
        to initialize the interpolation functions.
        If pre-computed triangulation not provided, then the constructor
        will calculate one.  This might take several minutes, so be patient.
        Much better to use pre-computed ones, as provided in, e.g.,
        :class:`dartmouth.Dartmouth_Isochrone`.
    :type tri: :class:`scipy.spatial.qhull.Delaunay`, optional

    :param minage,maxage:
        If desired, a minimum or maximum age can be manually entered.
        
    """
    def __init__(self,m_ini,age,feh,m_act,logL,Teff,logg,mags,tri=None,
                 minage=None, maxage=None, ext_table=False):
        """Warning: if tri object not provided, this will be very slow to be created.
        """

        self.minage = age.min()
        self.maxage = age.max()
        self.minmass = m_act.min()
        self.maxmass = m_act.max()
        self.minfeh = feh.min()
        self.maxfeh = feh.max()

        self.ext_table = ext_table

        if minage is not None:
            self.minage = minage
        if maxage is not None:
            self.maxage = maxage

        L = 10**logL

        if tri is None:
            points = np.zeros((len(m_ini),3))
            points[:,0] = m_ini
            points[:,1] = age
            points[:,2] = feh
            fn = interpnd(points,m_act)
            self.tri = fn.tri
        else:
            self.tri = tri
            self.mass = interpnd(self.tri,m_act)

        self._data = {'mass':m_act,
                    'logL':logL,
                    'logg':logg,
                    'logTeff':np.log10(Teff),
                    'mags':mags}
        self._props = ['mass', 'logL', 'logg', 'logTeff']

        self.bands = mags.keys()

        self._mag = {band:interpnd(self.tri,mags[band]) for band in self.bands}

        d = {}
        for b in self._mag.keys():
            d[b] = self._mag_fn(b)

        self.mag = d

    def _prop(self, prop, *args):
        if prop not in self._props:
            raise ValueError('Cannot call this function with {}.'.format(prop))
        attr = '_{}'.format(prop)
        if not hasattr(self, attr):
            setattr(self, attr, interpnd(self.tri, self._data[prop]))
        fn = getattr(self, attr)
        return fn(*args)

    def mass(self, *args):
        return self._prop('mass', *args)

    def logL(self, *args):
        return self._prop('logL', *args)

    def logg(self, *args):
        return self._prop('logg', *args)

    def logTeff(self, *args):
        return self._prop('logTeff', *args)

    def radius(self, *args):
        return np.sqrt(G*self.mass(*args)*MSUN/10**self.logg(*args))/RSUN

    def Teff(self, *args):
        return 10**self.logTeff(*args)

    def _mag_fn(self, band):
        def fn(mass, age, feh, distance=10, AV=0.0, x_ext=0., ext_table=self.ext_table):
            if x_ext==0.:
                ext = extcurve_0
            else:
                ext = extcurve(x_ext)
            if ext_table:
                A = AV*EXTINCTION[band]
            else:
                A = AV*ext(LAMBDA_EFF[band])
            dm = 5*np.log10(distance) - 5
            return self._mag[band](mass, age, feh) + dm + A
        return fn


    def __call__(self, mass, age, feh, 
                 distance=None, AV=0.0,
                 return_df=True, bands=None):
        """
        Returns all properties (or arrays of properties) at given mass, age, feh

        :param mass, age, feh:
            Mass, log(age), metallicity.  Can be float or array_like.

        :param distance:
            Distance in pc.  If passed, then mags will be converted to
            apparent mags based on distance (and ``AV``).

        :param AV:
            V-band extinction (magnitudes).

        :param return_df: (optional)
            If ``True``, return :class:``pandas.DataFrame`` containing all model
            parameters at each input value; if ``False``, return dictionary
            of the same.

        :param bands: (optional)
            List of photometric bands in which to return magnitudes.
            Must be subset of ``self.bands``.  If not set, then will
            default to returning all available bands. 

        :return:
            Either a :class:`pandas.DataFrame` or a dictionary containing
            model values evaluated at input points.
        """
        args = (mass, age, feh)
        Ms = self.mass(*args)*1
        Rs = self.radius(*args)*1
        logLs = self.logL(*args)*1
        loggs = self.logg(*args)*1
        Teffs = self.Teff(*args)*1
        if bands is None:
            bands = self.bands
        mags = {band:1*self.mag[band](*args) for band in bands}
        if distance is not None:
            dm = 5*np.log10(distance) - 5
            for band in mags:
                A = AV*EXTINCTION[band]
                mags[band] = mags[band] + dm + A
                
        
        props = {'age':age,'mass':Ms,'radius':Rs,'logL':logLs,
                'logg':loggs,'Teff':Teffs,'mag':mags}        

        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            try:
                df = pd.DataFrame(d)
            except ValueError:
                df = pd.DataFrame(d, index=[0])
            return df

    def agerange(self, m, feh=0.0):
        """
        For a given mass and feh, returns the min and max allowed ages.
        """
        ages = np.arange(self.minage, self.maxage, 0.01)
        rs = self.radius(m, ages, feh)
        w = np.where(np.isfinite(rs))[0]
        return ages[w[0]],ages[w[-1]]

    def evtrack(self,m,feh=0.0,minage=None,maxage=None,dage=0.02,
                return_df=True):
        """
        Returns evolution track for a single initial mass and feh.

        :param m: 
            Initial mass of desired evolution track.

        :param feh: (optional) 
            Metallicity of desired track.  Default = 0.0 (solar)

        :param minage, maxage: (optional)
            Minimum and maximum log(age) of desired track. Will default
            to min and max age of model isochrones. 

        :param dage: (optional)
            Spacing in log(age) at which to evaluate models.  Default = 0.02

        :param return_df: (optional)
            Whether to return a ``DataFrame`` or dicionary.  Default is ``True``.
            

        :return:
            Either a :class:`pandas.DataFrame` or dictionary
            representing the evolution
            track---fixed mass, sampled at chosen range of ages.
        
        """
        if minage is None:
            minage = self.minage
        if maxage is None:
            maxage = self.maxage
        ages = np.arange(minage,maxage,dage)
        Ms = self.mass(m,ages,feh)
        Rs = self.radius(m,ages,feh)
        logLs = self.logL(m,ages,feh)
        loggs = self.logg(m,ages,feh)
        Teffs = self.Teff(m,ages,feh)
        mags = {band:self.mag[band](m,ages,feh) for band in self.bands}

        props = {'age':ages,'mass':Ms,'radius':Rs,'logL':logLs,
                'logg':loggs, 'Teff':Teffs, 'mag':mags}

        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            try:
                df = pd.DataFrame(d)
            except ValueError:
                df = pd.DataFrame(d, index=[0])
            return df

            
    def isochrone(self,age,feh=0.0,minm=None,maxm=None,dm=0.02,
                  return_df=True,distance=None,AV=0.0):
        """
        Returns stellar models at constant age and feh, for a range of masses

        :param age: 
            log10(age) of desired isochrone.

        :param feh: (optional)
            Metallicity of desired isochrone (default = 0.0)

        :param minm, maxm: (optional)
            Mass range of desired isochrone (will default to max and min available)

        :param dm: (optional)
            Spacing in mass of desired isochrone.  Default = 0.02 Msun.

        :param return_df: (optional)
            Whether to return a :class:``pandas.DataFrame`` or dictionary.  Default is ``True``.
        
        :param distance:
            Distance in pc.  If passed, then mags will be converted to
            apparent mags based on distance (and ``AV``).

        :param AV:
            V-band extinction (magnitudes).            
        
        :return:
            :class:`pandas.DataFrame` or dictionary containing results.
        
        """
        if minm is None:
            minm = self.minmass
        if maxm is None:
            maxm = self.maxmass
        ms = np.arange(minm,maxm,dm)
        ages = np.ones(ms.shape)*age

        Ms = self.mass(ms,ages,feh)
        Rs = self.radius(ms,ages,feh)
        logLs = self.logL(ms,ages,feh)
        loggs = self.logg(ms,ages,feh)
        Teffs = self.Teff(ms,ages,feh)
        mags = {band:self.mag[band](ms,ages,feh) for band in self.bands}
        #for band in self.bands:
        #    mags[band] = self.mag[band](ms,ages)
        if distance is not None:
            dm = 5*np.log10(distance) - 5
            for band in mags:
                A = AV*EXTINCTION[band]
                mags[band] = mags[band] + dm + A

        props = {'M':Ms,'R':Rs,'logL':logLs,'logg':loggs,
                'Teff':Teffs,'mag':mags}        
        
        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            try:
                df = pd.DataFrame(d)
            except ValueError:
                df = pd.DataFrame(d, index=[0])
            return df
       
    def random_points(self,n,minmass=None,maxmass=None,
                      minage=None,maxage=None,
                      minfeh=None,maxfeh=None):
        """
        Returns n random mass, age, feh points, none of which are out of range.

        :param n:
            Number of desired points.

        :param minmass, maxmass: (optional)
            Desired allowed range.  Default is mass range of ``self``.

        :param minage, maxage: (optional)
            Desired allowed range.  Default is log10(age) range of
            ``self``.

        :param minfehs, maxfeh: (optional)
            Desired allowed range.  Default is feh range of ``self``.
                        
        :return:
            :class:`np.ndarray` arrays of randomly selected mass, log10(age),
            and feh values
            within allowed ranges.  Used, e.g., to initialize random walkers for
            :class:`StarModel` fits.

        ### Should change this to drawing from priors!
        """
        if minmass is None:
            minmass = self.minmass
        if maxmass is None:
            maxmass = self.maxmass
        if minage is None:
            minage = self.minage
        if maxage is None:
            maxage = self.maxage
        if minfeh is None:
            minfeh = self.minfeh
        if maxfeh is None:
            maxfeh = self.maxfeh

        ms = rand.uniform(minmass,maxmass,size=n)
        ages = rand.uniform(minage,maxage,size=n)
        fehs = rand.uniform(minage,maxage,size=n)

        Rs = self.radius(ms,ages,fehs)
        bad = np.isnan(Rs)
        nbad = bad.sum()
        while nbad > 0:
            ms[bad] = rand.uniform(minmass,maxmass,size=nbad)
            ages[bad] = rand.uniform(minage,maxage,size=nbad)
            fehs[bad] = rand.uniform(minfeh,maxfeh,size=nbad)
            Rs = self.radius(ms,ages,fehs)
            bad = np.isnan(Rs)
            nbad = bad.sum()
        return ms,ages,fehs


class MagFunction(object):
    def __init__(self, ic, band, icol):
        self.ic = ic
        self.band = band
        self.icol = icol
        self.x_ext = ic.x_ext
        self.ext_table = ic.ext_table

        if self.x_ext==0.:
            ext = extcurve_0
        else:
            ext = extcurve(x_ext)
        if self.ext_table:
            self.AAV = EXTINCTION[self.band]
        else:
            self.AAV = ext(LAMBDA_EFF[self.band])        

    def __call__(self, mass, age, feh, distance=10, AV=0.0, x_ext=None, ext_table=False):
        if x_ext is not None:
            if x_ext==0.:
                ext = extcurve_0
            else:
                ext = extcurve(x_ext)

            if ext_table:
                AAV = EXTINCTION[self.band]
            else:
                AAV = ext(LAMBDA_EFF[self.band])        
        else:
            AAV = self.AAV

        A = AV*AAV
        dm = 5*np.log10(distance) - 5
        mag = self.ic.interp_value(mass, age, feh, self.icol)
        return mag + dm + A

class LargeIsochrone(Isochrone):
    """Alternative isochrone implementation for large grids

    "large" means too large for Delaunay triangulation, as implemented in 
    :class:`Isochrone`.
    """
    name = 'default'
    modelgrid = ModelGrid
    age_col = 1
    feh_col = 7
    mass_col = 2
    loggTeff_col = 3
    logg_col = 4
    logL_col = 5

    def __init__(self, bands, x_ext=0., ext_table=False):
        # df should be indexed by [feh, age]

        self.df = self.modelgrid(bands).df
        self.bands = bands
        self.x_ext = 0.
        self.ext_table = ext_table

        self.Ncols = self.df.shape[1]
    
        self.fehs = self.df.iloc[:, self.feh_col].unique()
        self.ages = self.df.iloc[:, self.age_col].unique()
        self.Nfeh = len(self.fehs)
        self.Nage = len(self.ages)
    
        n_common_cols = len(self.modelgrid.common_columns)
        self._mag_cols = {b:n_common_cols+i for i,b in enumerate(self.bands)}
        # self._mag_cols = {'u':7, 'g':8, 'r':9, 'i':10, 'z':11}
        self.mag = {b: MagFunction(self, b, i)
                            for b,i in self._mag_cols.items()}
        # self.mag = {b:self._mag_fn(b) for b in self.bands}

        #organized array
        self._grid = None
        self._grid_Ns = None
        

    
        self.minage = self.ages.min()
        self.maxage = self.ages.max()
        self.minmass = self.df.iloc[:, self.mass_col].min()
        self.maxmass = self.df.iloc[:, self.mass_col].max()
        self.minfeh = self.fehs.min()
        self.maxfeh = self.fehs.max()

    def logTeff(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.loggTeff_col)

    def logg(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.logg_col)

    def logL(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.logL_col)

    def radius(self, *args):
        return np.sqrt(G*self.mass(*args)*MSUN/10**self.logg(*args))/RSUN

    def Teff(self, *args):
        return 10**self.logTeff(*args)

    def mass(self, *args):
        return args[0]

    @property
    def grid(self):
        if self._grid is None:
            self._make_grid()
        return self._grid
    
    @property
    def grid_Ns(self):
        if self._grid_Ns is None:
            self._make_grid()
        return self._grid_Ns
        
    @property
    def _npz_filename(self):
        return os.path.join(ISOCHRONES, self.name, '{}.npz'.format('-'.join(self.bands)))   

    def _make_grid(self, recalc=False):
        # Read from file if available.
        if os.path.exists(self._npz_filename) and not recalc:
            d = np.load(self._npz_filename)
            self._grid = d['grid']
            self._grid_Ns = d['grid_Ns']
        else:
            df_list = [[self.df.ix[f,a] for f in self.fehs] for a in self.ages]
            lens = np.array([[len(df_list[i][j]) for j in range(self.Nfeh)] 
                             for i in range(self.Nage)]).T #just because
            data = np.zeros((self.Nfeh, self.Nage, lens.max(), self.Ncols))

            for i in range(self.Nage):
                for j in range(self.Nfeh):
                    N = lens[j,i]
                    data[j, i, :N, :] = df_list[i][j].values
                    data[j, i, N:, :] = np.nan

            np.savez(self._npz_filename, grid=data, grid_Ns=lens)
            self._grid = data
            self._grid_Ns = lens
                
    def interp_value(self, mass, age, feh, icol): # 4 is log_g
        try:
            return interp_value(float(mass), float(age), float(feh), icol,
                                self.grid, self.mass_col,
                                self.ages, self.fehs, self.grid_Ns)

        except:
            # First, broadcast to common shape.
            b = np.broadcast(mass, age, feh)
            mass = np.resize(mass, b.shape).astype(float)
            age = np.resize(age, b.shape).astype(float)
            feh = np.resize(feh, b.shape).astype(float)

            # Then pass to helper function
            return interp_values(mass, age, feh, icol,
                                self.grid, self.mass_col,
                                self.ages, self.fehs, self.grid_Ns)




from astropy.coordinates import SkyCoord
import astropy.units as u

class EmptyQueryError(ValueError):
    pass

class Query(object):
    """ RA/dec in decimal degrees, pmra, pmdec in mas
    """
    def __init__(self, ra, dec, pmra=0., pmdec=0., epoch=2000., radius=5*u.arcsec):
        self.ra = ra
        self.dec = dec
        self.pmra = pmra
        self.pmdec = pmdec
        self.epoch = epoch

        if type(radius) in [type(1), type(1.)]:
            self.radius = radius*u.arcsec
        else:
            self.radius = radius

        self._coords = None

    def __str__(self):
        return '({0.ra}, {0.dec}), pm=({0.pmra}, {0.pmdec}), epoch={0.epoch}, radius={0.radius}'.format(self)

    def __repr__(self):
        return ('Query(ra={0.ra}, dec={0.dec}, pmra={0.pmra}, '.format(self) + 
                'pmdec={0.pmdec}, epoch={0.epoch}, radius={0.radius})'.format(self))

    @property
    def coords(self):
        if self._coords is None:
            self._coords = SkyCoord(self.ra, self.dec, unit='deg')
        return self._coords
    



import os, sys
import pandas as pd
import numpy as np
import logging

import emcee3
from emcee3.backends import Backend, HDFBackend

class Emcee3Model(emcee3.Model):
    def __init__(self, mod, *args, **kwargs):
        self.mod = mod
        super(Emcee3Model, self).__init__(*args, **kwargs)

    def compute_log_prior(self, state):
        state.log_prior = self.mod.lnprior(state.coords)
        return state

    def compute_log_likelihood(self, state):
        state.log_likelihood = self.mod.lnlike(state.coords)
        return state


def write_samples(mod, df, resultsdir='results'):
    """df is dataframe of samples, mod is model
    """

    if not os.path.exists(resultsdir):
        os.makedirs(resultsdir)
    samplefile = os.path.join(resultsdir, '{}.h5'.format(mod.name))
    df.to_hdf(samplefile, 'samples')

def fit_emcee3(mod, nwalkers=500, verbose=False, nsamples=5000, targetn=6,
                iter_chunksize=200, pool=None, overwrite=False,
                maxiter=10, sample_directory='mcmc_chains',
                nburn=3, mixedmoves=True, resultsdir='results', **kwargs):
    """fit model using Emcee3 

    modeled after https://github.com/dfm/gaia-kepler/blob/master/fit.py

    nburn is number of autocorr times to discard as burnin.
    """

    # Initialize
    walker = Emcee3Model(mod)
    ndim = mod.n_params

    if sample_directory is not None:
        sample_file = os.path.join(sample_directory, '{}.h5'.format(mod.name))
        if not os.path.exists(sample_directory):
            os.makedirs(sample_directory)
        backend = HDFBackend(sample_file)
        try:
            coords_init = backend.current_coords
        except (AttributeError, KeyError):
            coords_init = mod.sample_from_prior(nwalkers)
    else:
        backend = Backend()
        coords_init = mod.sample_from_prior(nwalkers)

    if mixedmoves:
        moves = [(emcee3.moves.KDEMove(), 0.4),
                 (emcee3.moves.DEMove(1.0), 0.4),
                 (emcee3.moves.DESnookerMove(), 0.2)]
    else:
        moves = emcee3.moves.KDEMove()

    sampler = emcee3.Sampler(moves, backend=backend)
    if overwrite:
        sampler.reset()
        coords_init = mod.emcee_p0(nwalkers)

    if pool is None:
        from emcee3.pools import DefaultPool
        pool = DefaultPool()

    ensemble = emcee3.Ensemble(walker, coords_init, pool=pool)

    def calc_stats(s):
        """returns tau_max, neff
        """
        tau = s.get_integrated_autocorr_time(c=1)
        tau_max = tau.max()
        neff = s.backend.niter / tau_max - nburn
        if verbose:
            print("Maximum autocorrelation time: {0}".format(tau_max))
            print("N_eff: {0} ({1})\n".format(neff * nwalkers, neff-nburn))            
        return tau_max, neff

    done = False
    if not overwrite:
        try:
            if verbose:
                print('Status from previous run:')
            tau_max, neff = calc_stats(sampler)
            if neff > targetn:
                done = True
        except (emcee3.autocorr.AutocorrError, KeyError):
            pass

    chunksize = iter_chunksize
    for iteration in range(maxiter):
        if done:
            break
        if verbose:
            print("Iteration {0}...".format(iteration + 1))
        sampler.run(ensemble, chunksize, progress=verbose)
        try:
            tau_max, neff = calc_stats(sampler)
        except emcee3.autocorr.AutocorrError:
            tau_max = 0
            continue
        if neff > targetn:
            done = True

    burnin = int(nburn*tau_max)
    ntot = nsamples
    samples = sampler.get_coords(flat=True, discard=burnin)
    total_samples = len(samples)
    if ntot > total_samples:
        ntot = total_samples
    if verbose:
        print("Discarding {0} samples for burn-in".format(burnin))
        print("Randomly choosing {0} samples".format(ntot))
    inds = np.random.choice(total_samples, size=ntot, replace=False)
    samples = samples[inds]

    df = pd.DataFrame(samples, columns=mod.param_names)
    write_samples(mod, df, resultsdir=resultsdir)
    
    return df
    # return sampler    


from __future__ import division,print_function
import os,os.path
import numpy as np
import pkg_resources
import logging

from scipy.interpolate import LinearNDInterpolator as interpnd
try:
    import pandas as pd
except ImportError:
    pd = None
    
import pickle

from .isochrone import Isochrone

#DATADIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))
DATADIR = os.getenv('ISOCHRONES',
                    os.path.expanduser(os.path.join('~','.isochrones')))
if not os.path.exists(DATADIR):
    os.mkdir(DATADIR)

MASTERFILE = '{}/parsec.h5'.format(DATADIR)
TRI_FILE = '{}/parsec.tri'.format(DATADIR)

####### utility function to generate master dataframe from raw files ######




from __future__ import print_function, division

import numpy as np
from ..extinction import extcurve

class Filter(object):
    lam_col = 0
    throughput_col = 1
    def __init__(self, filename, x=0):
        self.filename = filename
        self.x = x

        self._data = None
        self._extcurve = None
        self._extinction = None

    @property
    def data(self):
        if self._data is None:
            self._data = np.loadtxt(self.filename, usecols=(self.lam_col, 
                                                            self.throughput_col))
        return self._data
    

    @property
    def lam(self):
        return self.data[:, self.lam_col]
    
    @property
    def throughput(self):
        return self.data[:, self.throughput_col]

    @property
    def extcurve(self):
        if self._extcurve is None:
            self._extcurve = extcurve(self.x)
        return self._extcurve

    @property
    def extinction(self):
        if self._extinction is None:
            l, t, e = self.lam, self.throughput, self.extcurve(self.lam)
            self._extinction = np.trapz(t*e, l) / np.trapz(t, l)
        return self._extinction



from .config import on_rtd

import os, re, sys
import warnings
import logging

if not on_rtd:
    import pandas as pd
    import numpy as np
    from scipy.interpolate import LinearNDInterpolator as interpnd
    import numpy.random as rand
    import matplotlib.pyplot as plt

    from astropy import constants as const

    #Define useful constants
    G = const.G.cgs.value
    MSUN = const.M_sun.cgs.value
    RSUN = const.R_sun.cgs.value

    from .extinction import EXTINCTION, LAMBDA_EFF, extcurve, extcurve_0
    from .interp import interp_value, interp_values

else:
    G = 6.67e-11
    MSUN = 1.99e33
    RSUN = 6.96e10

from .config import ISOCHRONES
from .grid import ModelGrid

def get_ichrone(models, bands=None, default=False, **kwargs):
    """Gets Isochrone Object by name, or type, with the right bands

    If `default` is `True`, then will set bands
    to be the union of bands and default_bands
    """
    if isinstance(models, Isochrone):
        return models

    def actual(bands, ictype):
        if bands is None:
            return list(ictype.default_bands)
        elif default:
            return list(set(bands).union(set(ictype.default_bands)))
        else:
            return bands

    if type(models) is type(type):
        ichrone = models(actual(bands, models))
    elif models=='dartmouth':
        from isochrones.dartmouth import Dartmouth_Isochrone
        ichrone = Dartmouth_Isochrone(bands=actual(bands, Dartmouth_Isochrone), **kwargs)
    elif models=='dartmouthfast':
        from isochrones.dartmouth import Dartmouth_FastIsochrone
        ichrone = Dartmouth_FastIsochrone(bands=actual(bands, Dartmouth_FastIsochrone), **kwargs)
    elif models=='mist':
        from isochrones.mist import MIST_Isochrone
        ichrone = MIST_Isochrone(bands=actual(bands, MIST_Isochrone), **kwargs)
    elif models=='padova':
        from isochrones.padova import Padova_Isochrone
        ichrone = Padova_Isochrone(bands=actual(bands, Padova_Isochrone), **kwargs)
    elif models=='basti':
        from isochrones.basti import Basti_Isochrone
        ichrone = Basti_Isochrone(bands=actual(bands, Basti_Isochrone), **kwargs)
    else:
        raise ValueError('Unknown stellar models: {}'.format(models))
    return ichrone


class Isochrone(object):
    """
    Basic isochrone class. Everything is a function of mass, log(age), Fe/H.

    Can be instantiated directly, but will typically be used with a pre-defined
    subclass, such as :class:`dartmouth.Dartmouth_Isochrone`.  All parameters
    must be array-like objects of the same length, with the exception of ``mags``,
    which is a dictionary of such array-like objects.

    :param m_ini:
        Array of initial mass values [msun].
    :type m_ini: array-like

    :param age:
        log10(age) [yr]

    :param feh:
        Metallicity [dex]

    :param m_act:
        Actual mass; same as m_ini if mass loss not implemented [msun]

    :param logL:
        log10(luminosity) [solar units]

    :param Teff:
        Effective temperature [K]

    :param logg:
        log10(surface gravity) [cgs]

    :param mags:
        Dictionary of absolute magnitudes in different bands
    :type mags: ``dict``

    :param tri:
        Triangulation object used
        to initialize the interpolation functions.
        If pre-computed triangulation not provided, then the constructor
        will calculate one.  This might take several minutes, so be patient.
        Much better to use pre-computed ones, as provided in, e.g.,
        :class:`dartmouth.Dartmouth_Isochrone`.
    :type tri: :class:`scipy.spatial.qhull.Delaunay`, optional

    :param minage,maxage:
        If desired, a minimum or maximum age can be manually entered.

    """
    def __init__(self,m_ini,age,feh,m_act,logL,Teff,logg,mags,tri=None,
                 minage=None, maxage=None, ext_table=False):
        """Warning: if tri object not provided, this will be very slow to be created.
        """

        self.minage = age.min()
        self.maxage = age.max()
        self.minmass = m_act.min()
        self.maxmass = m_act.max()
        self.minfeh = feh.min()
        self.maxfeh = feh.max()

        self.ext_table = ext_table

        if minage is not None:
            logging.warning("minage and maxage keywords are deprecated." + \
                          "Use instead the .set_bounds(age=(lo, hi)) attribute of StarModel.")
            self.minage = minage
        if maxage is not None:
            logging.warning("minage and maxage keywords are deprecated." + \
                          "Use instead the .set_bounds(age=(lo, hi)) attribute of StarModel.")
            self.maxage = maxage

        L = 10**logL

        if tri is None:
            points = np.zeros((len(m_ini),3))
            points[:,0] = m_ini
            points[:,1] = age
            points[:,2] = feh
            fn = interpnd(points,m_act)
            self.tri = fn.tri
        else:
            self.tri = tri
            self.mass = interpnd(self.tri,m_act)

        self._data = {'mass':m_act,
                    'logL':logL,
                    'logg':logg,
                    'logTeff':np.log10(Teff),
                    'mags':mags}
        self._props = ['mass', 'logL', 'logg', 'logTeff']

        self.bands = list(mags.keys())

        self._mag = {band:interpnd(self.tri,mags[band]) for band in self.bands}

        self.mag = {b : self._mag_fn(b) for b in self.bands}

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['mag'] # This can't be pickled
        return odict

    def __setstate__(self, odict):
        self.__dict__ = odict
        self.__dict__['mag'] = {b : self._mag_fn(b) for b in self.bands}

    def _prop(self, prop, *args):
        if prop not in self._props:
            raise ValueError('Cannot call this function with {}.'.format(prop))
        attr = '_{}'.format(prop)
        if not hasattr(self, attr):
            setattr(self, attr, interpnd(self.tri, self._data[prop]))
        fn = getattr(self, attr)
        return fn(*args)

    def mass(self, *args):
        return self._prop('mass', *args)

    def logL(self, *args):
        return self._prop('logL', *args)

    def logg(self, *args):
        return self._prop('logg', *args)

    def logTeff(self, *args):
        return self._prop('logTeff', *args)

    def radius(self, *args):
        return np.sqrt(G*self.mass(*args)*MSUN/10**self.logg(*args))/RSUN

    def Teff(self, *args):
        return 10**self.logTeff(*args)

    def density(self, *args):
        """ Mean density in g/cc
        """
        M = self.mass(*args) * MSUN
        V = 4./3 * np.pi * (self.radius(*args) * RSUN)**3
        return  M/V

    def delta_nu(self, *args):
        """Returns asteroseismic delta_nu in uHz

        reference: https://arxiv.org/pdf/1312.3853v1.pdf, Eq (2)
        """
        return 134.88 * np.sqrt(self.mass(*args) / self.radius(*args)**3)

    def nu_max(self, *args):
        """Returns asteroseismic nu_max in uHz

        reference: https://arxiv.org/pdf/1312.3853v1.pdf, Eq (3)
        """
        return 3120.* (self.mass(*args) /
                        (self.radius(*args)**2 * np.sqrt(self.Teff(*args)/5777.)))

    def _mag_fn(self, band):
        def fn(mass, age, feh, distance=10, AV=0.0, x_ext=0., ext_table=self.ext_table):
            if x_ext==0.:
                ext = extcurve_0
            else:
                ext = extcurve(x_ext)
            if ext_table:
                A = AV*EXTINCTION[band]
            else:
                A = AV*ext(LAMBDA_EFF[band])
            dm = 5*np.log10(distance) - 5
            return self._mag[band](mass, age, feh) + dm + A
        return fn


    def __call__(self, mass, age, feh,
                 distance=None, AV=0.0,
                 return_df=True, bands=None):
        """
        Returns all properties (or arrays of properties) at given mass, age, feh

        :param mass, age, feh:
            Mass, log(age), metallicity.  Can be float or array_like.

        :param distance:
            Distance in pc.  If passed, then mags will be converted to
            apparent mags based on distance (and ``AV``).

        :param AV:
            V-band extinction (magnitudes).

        :param return_df: (optional)
            If ``True``, return :class:``pandas.DataFrame`` containing all model
            parameters at each input value; if ``False``, return dictionary
            of the same.

        :param bands: (optional)
            List of photometric bands in which to return magnitudes.
            Must be subset of ``self.bands``.  If not set, then will
            default to returning all available bands.

        :return:
            Either a :class:`pandas.DataFrame` or a dictionary containing
            model values evaluated at input points.
        """
        # Broadcast inputs to the same shape
        mass, age, feh = [np.array(a) for a in np.broadcast_arrays(mass, age, feh)]
        args = (mass, age, feh)
        Ms = self.mass(*args)*1
        Rs = self.radius(*args)*1
        logLs = self.logL(*args)*1
        loggs = self.logg(*args)*1
        Teffs = self.Teff(*args)*1
        if bands is None:
            bands = self.bands
        if distance is not None:
            args += (distance, AV)
        mags = {band:1*self.mag[band](*args) for band in bands}
        # if distance is not None:
        #     dm = 5*np.log10(distance) - 5
        #     for band in mags:
        #         A = AV*EXTINCTION[band]
        #         mags[band] = mags[band] + dm + A


        props = {'age':age,'mass':Ms,'radius':Rs,'logL':logLs,
                'logg':loggs,'Teff':Teffs,'mag':mags}

        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            df = pd.DataFrame(d)
            return df

    def agerange(self, m, feh=0.0):
        """
        For a given mass and feh, returns the min and max allowed ages.
        """
        ages = np.arange(self.minage, self.maxage, 0.01)
        rs = self.radius(m, ages, feh)
        w = np.where(np.isfinite(rs))[0]
        return ages[w[0]],ages[w[-1]]

    def evtrack(self,m,feh=0.0,minage=None,maxage=None,dage=0.02,
                return_df=True):
        """
        Returns evolution track for a single initial mass and feh.

        :param m:
            Initial mass of desired evolution track.

        :param feh: (optional)
            Metallicity of desired track.  Default = 0.0 (solar)

        :param minage, maxage: (optional)
            Minimum and maximum log(age) of desired track. Will default
            to min and max age of model isochrones.

        :param dage: (optional)
            Spacing in log(age) at which to evaluate models.  Default = 0.02

        :param return_df: (optional)
            Whether to return a ``DataFrame`` or dicionary.  Default is ``True``.


        :return:
            Either a :class:`pandas.DataFrame` or dictionary
            representing the evolution
            track---fixed mass, sampled at chosen range of ages.

        """
        if minage is None:
            minage = self.minage
        if maxage is None:
            maxage = self.maxage
        ages = np.arange(minage,maxage,dage)
        Ms = self.mass(m,ages,feh)
        Rs = self.radius(m,ages,feh)
        logLs = self.logL(m,ages,feh)
        loggs = self.logg(m,ages,feh)
        Teffs = self.Teff(m,ages,feh)
        mags = {band:self.mag[band](m,ages,feh) for band in self.bands}

        props = {'age':ages,'mass':Ms,'radius':Rs,'logL':logLs,
                'logg':loggs, 'Teff':Teffs, 'mag':mags}

        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            try:
                df = pd.DataFrame(d)
            except ValueError:
                df = pd.DataFrame(d, index=[0])
            return df


    def isochrone(self,age,feh=0.0,minm=None,maxm=None,dm=0.02,
                  return_df=True,distance=None,AV=0.0):
        """
        Returns stellar models at constant age and feh, for a range of masses

        :param age:
            log10(age) of desired isochrone.

        :param feh: (optional)
            Metallicity of desired isochrone (default = 0.0)

        :param minm, maxm: (optional)
            Mass range of desired isochrone (will default to max and min available)

        :param dm: (optional)
            Spacing in mass of desired isochrone.  Default = 0.02 Msun.

        :param return_df: (optional)
            Whether to return a :class:``pandas.DataFrame`` or dictionary.  Default is ``True``.

        :param distance:
            Distance in pc.  If passed, then mags will be converted to
            apparent mags based on distance (and ``AV``).

        :param AV:
            V-band extinction (magnitudes).

        :return:
            :class:`pandas.DataFrame` or dictionary containing results.

        """
        if minm is None:
            minm = self.minmass
        if maxm is None:
            maxm = self.maxmass
        ms = np.arange(minm,maxm,dm)
        ages = np.ones(ms.shape)*age

        Ms = self.mass(ms,ages,feh)
        Rs = self.radius(ms,ages,feh)
        logLs = self.logL(ms,ages,feh)
        loggs = self.logg(ms,ages,feh)
        Teffs = self.Teff(ms,ages,feh)
        mags = {band:self.mag[band](ms,ages,feh) for band in self.bands}
        #for band in self.bands:
        #    mags[band] = self.mag[band](ms,ages)
        if distance is not None:
            dm = 5*np.log10(distance) - 5
            for band in mags:
                A = AV*EXTINCTION[band]
                mags[band] = mags[band] + dm + A

        props = {'M':Ms,'R':Rs,'logL':logLs,'logg':loggs,
                'Teff':Teffs,'mag':mags}

        if not return_df:
            return props
        else:
            d = {}
            for key in props.keys():
                if key=='mag':
                    for m in props['mag'].keys():
                        d['{}_mag'.format(m)] = props['mag'][m]
                else:
                    d[key] = props[key]
            try:
                df = pd.DataFrame(d)
            except ValueError:
                df = pd.DataFrame(d, index=[0])
            return df

    def random_points(self,n,minmass=None,maxmass=None,
                      minage=None,maxage=None,
                      minfeh=None,maxfeh=None):
        """
        Returns n random mass, age, feh points, none of which are out of range.

        :param n:
            Number of desired points.

        :param minmass, maxmass: (optional)
            Desired allowed range.  Default is mass range of ``self``.

        :param minage, maxage: (optional)
            Desired allowed range.  Default is log10(age) range of
            ``self``.

        :param minfehs, maxfeh: (optional)
            Desired allowed range.  Default is feh range of ``self``.

        :return:
            :class:`np.ndarray` arrays of randomly selected mass, log10(age),
            and feh values
            within allowed ranges.  Used, e.g., to initialize random walkers for
            :class:`StarModel` fits.

        .. todo::

            Should change this to drawing from priors!  Current implementation
            is a bit outdated.
        """
        if minmass is None:
            minmass = self.minmass
        if maxmass is None:
            maxmass = self.maxmass
        if minage is None:
            minage = self.minage
        if maxage is None:
            maxage = self.maxage
        if minfeh is None:
            minfeh = self.minfeh
        if maxfeh is None:
            maxfeh = self.maxfeh

        ms = rand.uniform(minmass,maxmass,size=n)
        ages = rand.uniform(minage,maxage,size=n)
        fehs = rand.uniform(minage,maxage,size=n)

        Rs = self.radius(ms,ages,fehs)
        bad = np.isnan(Rs)
        nbad = bad.sum()
        while nbad > 0:
            ms[bad] = rand.uniform(minmass,maxmass,size=nbad)
            ages[bad] = rand.uniform(minage,maxage,size=nbad)
            fehs[bad] = rand.uniform(minfeh,maxfeh,size=nbad)
            Rs = self.radius(ms,ages,fehs)
            bad = np.isnan(Rs)
            nbad = bad.sum()
        return ms,ages,fehs


class MagFunction(object):
    def __init__(self, ic, band, icol):
        self.ic = ic
        self.band = band
        self.icol = icol
        self.x_ext = ic.x_ext
        self.ext_table = ic.ext_table

        if self.x_ext==0.:
            ext = extcurve_0
        else:
            ext = extcurve(x_ext)
        if self.ext_table:
            self.AAV = EXTINCTION[self.band]
        else:
            self.AAV = ext(LAMBDA_EFF[self.band])

    def __call__(self, mass, age, feh, distance=10, AV=0.0, x_ext=None, ext_table=False):
        if x_ext is not None:
            if x_ext==0.:
                ext = extcurve_0
            else:
                ext = extcurve(x_ext)

            if ext_table:
                AAV = EXTINCTION[self.band]
            else:
                AAV = ext(LAMBDA_EFF[self.band])
        else:
            AAV = self.AAV

        A = AV*AAV
        dm = 5*np.log10(distance) - 5
        mag = self.ic.interp_value(mass, age, feh, self.icol)
        return mag + dm + A

class FastIsochrone(Isochrone):
    """Alternative isochrone implementation for large grids, faster likelihoods

    This implementation allows faster point interpolations than
    the triangulation method in the base :class:`Isochrone` class;
    however, for simulating large populations (passing arrays of parameters
    at a time), the base :class:`Isochrone` is still faster.  This is
    also the go-to implementation for grids that are too large to make the
    Delaunay method feasible (e.g., significantly over 100,000 points in the grid.)

    However, edge effects might be a bit more problematic here than in the base class,
    meaning fitting very evolved stars might be an issue. Also, don't trust this subclass
    for the :function:`Isochrone.agerange` function, for the same reason.

    Subclasses must set the appropriate attributes, and then things should work.
    """
    name = 'default'
    modelgrid = ModelGrid
    age_col = None
    feh_col = None
    mass_col = None
    loggTeff_col = None
    logg_col = None
    logL_col = None
    default_bands = ('g')

    def __init__(self, bands=None, x_ext=0., ext_table=False, debug=False, **kwargs):
        # df should be indexed by [feh, age]

        if bands is None:
            bands = list(self.default_bands)
        bands = sorted(bands)
        self.bands = bands

        self._df = None
        self.x_ext = 0.
        self.ext_table = ext_table
        self.debug = debug


        self._fehs = None
        self._ages = None
        self._Nfeh = None
        self._Nage = None

        self._minage = None
        self._maxage = None
        self._minmass = None
        self._maxmass = None
        self._minfeh = None
        self._maxfeh = None

        n_common_cols = len(self.modelgrid.get_common_columns(**kwargs))
        self._mag_cols = {b:n_common_cols+i for i,b in enumerate(self.bands)}
        self.mag = {b: MagFunction(self, b, i)
                            for b,i in self._mag_cols.items()}

        #organized array
        self._grid = None
        self._grid_Ns = None

        # kwargs to pass to self.modelgrid
        self.modelgrid_kwargs = kwargs

    def _initialize(self):
        for attr in ['df','Ncols','fehs','ages','Nfeh','Nage',
                     'minage','maxage','minfeh','maxfeh','minmass','maxmass']:
            _ = getattr(self, attr)


    @property
    def df(self):
        if self._df is None:
            self._df = self.modelgrid(self.bands, **self.modelgrid_kwargs).df
        return self._df

    @property
    def Ncols(self):
        return self.df.shape[1]

    @property
    def fehs(self):
        if self._fehs is None:
            self._fehs = self.df.iloc[:, self.feh_col].unique()
        return self._fehs

    @property
    def ages(self):
        if self._ages is None:
            self._ages = self.df.iloc[:, self.age_col].unique()
        return self._ages

    @property
    def Nfeh(self):
        if self._Nfeh is None:
            self._Nfeh = len(self.fehs)
        return self._Nfeh

    @property
    def Nage(self):
        if self._Nage is None:
            self._Nage = len(self.ages)
        return self._Nage

    @property
    def minage(self):
        if self._minage is None:
            self._minage = self.ages.min()
        return self._minage

    @property
    def maxage(self):
        if self._maxage is None:
            self._maxage = self.ages.max()
        return self._maxage

    @property
    def minfeh(self):
        if self._minfeh is None:
            self._minfeh = self.fehs.min()
        return self._minfeh

    @property
    def maxfeh(self):
        if self._maxfeh is None:
            self._maxfeh = self.fehs.max()
        return self._maxfeh

    @property
    def minmass(self):
        if self._minmass is None:
            self._minmass = self.df.iloc[:, self.mass_col].min()
        return self._minmass

    @property
    def maxmass(self):
        if self._maxmass is None:
            self._maxmass = self.df.iloc[:, self.mass_col].max()
        return self._maxmass

    def logTeff(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.loggTeff_col)

    def logg(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.logg_col)

    def logL(self, mass, age, feh):
        return self.interp_value(mass, age, feh, self.logL_col)

    def mass(self, *args):
        if np.size(args[0]) > 1:
            return np.array(args[0])
        else:
            return args[0]

    @property
    def grid(self):
        if self._grid is None:
            self._make_grid()
        return self._grid

    @property
    def grid_Ns(self):
        if self._grid_Ns is None:
            self._make_grid()
        return self._grid_Ns

    @property
    def _npz_filename(self):
        keys = list(self.modelgrid_kwargs.keys())
        keys.sort()

        filename = os.path.join(ISOCHRONES, self.name, '{}'.format('-'.join(self.bands)))

        for k in keys:
            filename += '_{}{}'.format(k, self.modelgrid_kwargs[k])

        filename += '.npz'
        return filename

    def _make_grid(self, recalc=False):
        # Read from file if available.
        if os.path.exists(self._npz_filename) and not recalc:
            d = np.load(self._npz_filename)
            self._grid = d['grid']
            self._grid_Ns = d['grid_Ns']
        else:
            df_list = [[self.df.ix[f,a] for f in self.fehs] for a in self.ages]
            lens = np.array([[len(df_list[i][j]) for j in range(self.Nfeh)]
                             for i in range(self.Nage)]).T #just because
            data = np.zeros((self.Nfeh, self.Nage, lens.max(), self.Ncols))

            for i in range(self.Nage):
                for j in range(self.Nfeh):
                    N = lens[j,i]
                    data[j, i, :N, :] = df_list[i][j].values
                    data[j, i, N:, :] = np.nan

            np.savez(self._npz_filename, grid=data, grid_Ns=lens)
            self._grid = data
            self._grid_Ns = lens

    def interp_value(self, mass, age, feh, icol): # 4 is log_g
        if self._ages is None:
            self._initialize()

        try:
            return interp_value(float(mass), float(age), float(feh), icol,
                                self.grid, self.mass_col,
                                self.ages, self.fehs, self.grid_Ns, self.debug)

        except:
            # First, broadcast to common shape.
            b = np.broadcast(mass, age, feh)
            mass = np.resize(mass, b.shape).astype(float)
            age = np.resize(age, b.shape).astype(float)
            feh = np.resize(feh, b.shape).astype(float)

            # Then pass to helper function
            return interp_values(mass, age, feh, icol,
                                self.grid, self.mass_col,
                                self.ages, self.fehs, self.grid_Ns)



__version__ = '1.2.1'

try:
    __ISOCHRONES_SETUP__
except NameError:
    __ISOCHRONES_SETUP__ = False

if not __ISOCHRONES_SETUP__:
    __all__ = ['get_ichrone', 'Isochrone', 'StarModel', 'ModelGrid']
    from .isochrone import Isochrone, get_ichrone
    from .starmodel import StarModel, BinaryStarModel, TripleStarModel
    from .grid import ModelGrid



from __future__ import print_function, division
import logging 

from .config import on_rtd

if not on_rtd:
    import numpy as np
    from scipy.stats import uniform 
    from scipy.integrate import quad
    import matplotlib.pyplot as plt

class Prior(object):

    def __call__(self, x, **kwargs):
        return self.pdf(x, **kwargs)

    def pdf(self, x):
        raise NotImplementedError

    def sample(self, n):
        raise NotImplementedError

    def test_integral(self):
        assert np.isclose(1, quad(self.pdf, -np.inf, np.inf)[0])

    def test_sampling(self, n=100000, plot=False):
        x = self.sample(n)
        if hasattr(self, 'bounds'):
            rng = self.bounds
        else:
            rng = None
        hn, _ = np.histogram(x, range=rng)
        h, b = np.histogram(x, density=True, range=rng)
        logging.debug('bins: {}'.format(b))
        logging.debug('raw: {}'.format(hn))
        pdf = np.array([quad(self.pdf,lo,hi)[0]/(hi-lo) for lo,hi in zip(b[:-1], b[1:])])
        sigma = 1./np.sqrt(hn)
        resid = np.absolute(pdf - h) / pdf
        logging.debug('pdf: {}'.format(pdf))
        logging.debug('hist: {}'.format(h))
        logging.debug('sigma: {}'.format(sigma))
        logging.debug('{}'.format(resid/sigma))

        c = (b[:-1] + b[1:])/2
        if plot:
            plt.plot(c, h, '_')
            plt.plot(c, pdf, 'x')
        else:
            assert max((resid / sigma)[hn > 50]) < 4

class BoundedPrior(Prior):
    def __init__(self, bounds=None):
        self.bounds = bounds
        super(BoundedPrior, self).__init__()

    def test_integral(self):
        assert np.isclose(1, quad(self.pdf, *self.bounds)[0])

    def __call__(self, x):
        lo, hi = self.bounds
        if x < lo or x > hi:
            return 0
        else:
            return self.pdf(x)

class FlatPrior(BoundedPrior):

    def __init__(self, bounds):
        super(FlatPrior, self).__init__(bounds=bounds)

    def pdf(self, x):
        lo, hi = self.bounds
        return 1./(hi - lo)

    def sample(self, n):
        lo, hi = self.bounds
        return np.random.random(n)*(hi - lo) + lo

class FlatLogPrior(BoundedPrior):
    def __init__(self, bounds):
        super(FlatLogPrior, self).__init__(bounds=bounds)

    def pdf(self, x):
        lo, hi = self.bounds
        return np.log(10) * 10**x/ (10**hi - 10**lo)

    def sample(self, n):
        lo, hi = self.bounds
        return np.log10(np.random.random(n)*(10**hi - 10**lo) + 10**lo)

class PowerLawPrior(BoundedPrior):
    def __init__(self, alpha, bounds=None):
        self.alpha = alpha
        super(PowerLawPrior, self).__init__(bounds=bounds)

    def pdf(self, x):
        lo, hi = self.bounds
        C = (1 + self.alpha)/(hi**(1 + self.alpha) - lo**(1 + self.alpha))
        # C = 1/(1/(self.alpha+1)*(1 - lo**(self.alpha+1)))
        return C * x**self.alpha

    def sample(self, n):
        """

        cdf = C * int(x**a)|x=lo..x

            = C * [x**(a+1) / (a+1)] | x=lo..x
            = C * ([x**(a+1) / (a+1)] - [lo**(a+1) / (a+1)])
         u  =
         
         u/C + [lo**(a+1) / (a+1)] = x**(a+1) / (a+1)
         (a+1) * (u/C + [lo**(a+1) / (a+1)]) = x**(a+1)
         [(a+1) * (u/C + [lo**(a+1) / (a+1)])] ** (1/(a+1)) = x 
        """
        lo, hi = self.bounds
        C = (1 + self.alpha)/(hi**(1 + self.alpha) - lo**(1 + self.alpha))
        u = np.random.random(n)
        a = self.alpha
        return ((a+1) * (u/C + (lo**(a+1) / (a+1))))**(1/(a+1))        

class FehPrior(Prior):
    """feh PDF based on local SDSS distribution
    
    From Jo Bovy:
    https://github.com/jobovy/apogee/blob/master/apogee/util/__init__.py#L3
    2D gaussian fit based on Casagrande (2011)
    """

    def __init__(self, halo_fraction=0.001):   
        self.halo_fraction = halo_fraction
        super(FehPrior, self).__init__()

    def pdf(self, x):
        feh = x

        disk_norm = 2.5066282746310007 # integral of the below from -np.inf to np.inf
        disk_fehdist = 1./disk_norm * (0.8/0.15*np.exp(-0.5*(feh-0.016)**2./0.15**2.)\
                       +0.2/0.22*np.exp(-0.5*(feh+0.15)**2./0.22**2.))

        halo_mu, halo_sig = -1.5, 0.4
        halo_fehdist = 1./np.sqrt(2*np.pi*halo_sig**2) * np.exp(-0.5*(feh-halo_mu)**2/halo_sig**2)

        return self.halo_fraction * halo_fehdist + (1 - self.halo_fraction)*disk_fehdist

    def sample(self, n):
        w1, mu1, sig1 = 0.8, 0.016, 0.15
        w2, mu2, sig2 = 0.2, -0.15, 0.22

        halo_mu, halo_sig = -1.5, 0.4

        x1 = np.random.randn(n)*sig1 + mu1
        x2 = np.random.randn(n)*sig2 + mu2

        xhalo = np.random.randn(n)*halo_sig + halo_mu

        u1 = np.random.random(n)
        x = x1
        m1 = u1 < w2
        x[m1] = x2[m1]

        u2 = np.random.random(n)
        m2 = u2 < self.halo_fraction
        x[m2] = xhalo[m2]
        return x

#  Uniform true age prior; where 'age' is actually log10(age)
age_prior = FlatLogPrior(bounds=(9, 10.15))
distance_prior = PowerLawPrior(alpha=2., bounds=(0,3000))
AV_prior = FlatPrior(bounds=(0., 1.))
q_prior = PowerLawPrior(alpha=0.3, bounds=(0.1, 1))
salpeter_prior = PowerLawPrior(alpha=-2.35, bounds=(0.1, 10))
feh_prior = FehPrior(halo_fraction=0.001)




import os,re, glob
import numpy as np
import pandas as pd
import logging

from ..config import ISOCHRONES
from ..grid import ModelGrid

class DartmouthModelGrid(ModelGrid):
    """Grid of Dartmouth Models.

    The following photometric systems are included::

        phot_systems = ('SDSSugriz','UBVRIJHKsKp','WISE','LSST','UKIDSS')
        phot_bands = dict(SDSSugriz=['sdss_z', 'sdss_i', 'sdss_r', 'sdss_u', 'sdss_g'],
              UBVRIJHKsKp=['B', 'I', 'H', 'J', 'Ks', 'R', 'U', 'V', 'D51', 'Kp'],
              WISE=['W4', 'W3', 'W2', 'W1'],
              LSST=['LSST_r', 'LSST_u', 'LSST_y', 'LSST_z', 'LSST_g', 'LSST_i'],
              UKIDSS=['Y', 'H', 'K', 'J', 'Z'])

    You may add additional ones by putting the ``phot_system.tgz`` file
    into ``.isochrones/dartmouth`` alongside the others.  If you do
    this, you must edit the object definition accordingly so the ``get_band``
    function returns the correct information.
    """
    name = 'dartmouth'
    common_columns = ('EEP', 'MMo', 'LogTeff', 'LogG', 'LogLLo', 'age', 'feh')
    phot_systems = ('SDSSugriz','UBVRIJHKsKp','WISE','LSST','UKIDSS','HST_WFC3')
    phot_bands = dict(SDSSugriz=['sdss_z', 'sdss_i', 'sdss_r', 'sdss_u', 'sdss_g'],
                  UBVRIJHKsKp=['B', 'I', 'H', 'J', 'Ks', 'R', 'U', 'V', 'D51', 'Kp'],
                  WISE=['W4', 'W3', 'W2', 'W1'],
                  LSST=['LSST_r', 'LSST_u', 'LSST_y', 'LSST_z', 'LSST_g', 'LSST_i'],
                  UKIDSS=['Y', 'H', 'K', 'J', 'Z'],
                  HST_WFC3=['uvf200l', 'uvf218w', 'uvf225w', 'uvf275w', 'uvf280n', 'uvf300x',
                            'uvf336w', 'uvf343n', 'uvf350l', 'uvf373n', 'uvf390m', 'uvf390w',
                            'uvf395n', 'uvf410m', 'uvf438w', 'uvf467m', 'uvf469n', 'uvf475w',
                            'uvf475x', 'uvf487n', 'uvf502n', 'uvf547m', 'uvf555w', 'uvf600l',
                            'uvf606w', 'uvf621m', 'uvf625w', 'uvf631n', 'uvf645n', 'uvf656n',
                            'uvf657n', 'uvf658n', 'uvf665n', 'uvf673n', 'uvf680n', 'uvf689m',
                            'uvf763m', 'uvf775w', 'uvf814w', 'uvf845m', 'uvf850l', 'uvf953n',
                            'uvfq232', 'uvfq243', 'uvfq378', 'uvfq387', 'uvfq422', 'uvfq436',
                            'uvfq437', 'uvfq492', 'uvfq508', 'uvfq575', 'uvfq619', 'uvfq634',
                            'uvfq672', 'uvfq674', 'uvfq727', 'uvfq750', 'uvfq889', 'uvfq906',
                            'uvfq924', 'uvfq937', 'irf098m', 'irf105w', 'irf110w', 'irf125w',
                            'irf126n', 'irf127m', 'irf128n', 'irf130n', 'irf132n', 'irf139m',
                            'irf140w', 'irf153m', 'irf160w', 'irf164n', 'irf167n'])

    default_kwargs = {'afe':'afep0', 'y':''}
    datadir = os.path.join(ISOCHRONES, 'dartmouth')
    zenodo_record = 1002927
    zenodo_files = ('dartmouth.tri',)
    zenodo_md5 = ('570b758ea98c8a5a806149bd1b854b98',)
    master_tarball_file = 'dartmouth.tgz'

    default_bands = ('B','V','g','r','i','z',
                     'J','H','K',
                     'W1','W2','W3','Kepler')

    extra_url_base = 'http://stellar.dartmouth.edu/models/isochrones'

    @classmethod
    def get_band(cls, b, **kwargs):
        """Defines what a "shortcut" band name refers to.

        """
        phot = None

        # Default to SDSS for these
        if b in ['u','g','r','i','z']:
            phot = 'SDSSugriz'
            band = 'sdss_{}'.format(b)
        elif b in ['U','B','V','R','I','J','H','Ks']:
            phot = 'UBVRIJHKsKp'
            band = b
        elif b=='K':
            phot = 'UBVRIJHKsKp'
            band = 'Ks'
        elif b in ['kep','Kepler','Kp']:
            phot = 'UBVRIJHKsKp'
            band = 'Kp'
        elif b in ['W1','W2','W3','W4']:
            phot = 'WISE'
            band = b
        elif re.match('uvf', b) or re.match('irf', b):
            phot = 'HST_WFC3'
            band = b
        else:
            m = re.match('([a-zA-Z]+)_([a-zA-Z_]+)',b)
            if m:
                if m.group(1) in cls.phot_systems:
                    phot = m.group(1)
                    if phot=='LSST':
                        band = b
                    else:
                        band = m.group(2)
                elif m.group(1) in ['UK','UKIRT']:
                    phot = 'UKIDSS'
                    band = m.group(2)
        if phot is None:
            raise ValueError('Dartmouth Models cannot resolve band {}!'.format(b))
        return phot, band

    @classmethod
    def phot_tarball_file(cls, phot, **kwargs):
        return os.path.join(cls.datadir, '{}.tgz'.format(phot))

    def get_filenames(self, phot, afe='afep0', y=''):
        if not os.path.exists(os.path.join(self.datadir, 'isochrones', phot)):
            if not os.path.exists(self.phot_tarball_file(phot, afe=afe, y=y)):
                self.extract_phot_tarball(phot)
                # self.extract_master_tarball()
            # self.extract_phot_tarball(phot)

        return glob.glob('{3}/isochrones/{0}/*{1}{2}.{0}*'.format(phot,afe,y,self.datadir))

    @classmethod
    def get_feh(cls, filename):
        m = re.search('feh([mp])(\d+)afe', filename)
        if m:
            sign = 1 if m.group(1)=='p' else -1
            return float(m.group(2))/10 * sign

    @classmethod
    def to_df(cls, filename):
        try:
            rec = np.recfromtxt(filename,skip_header=8,names=True)
        except:
            print('Error reading {}!'.format(filename))
            raise RuntimeError
        df = pd.DataFrame(rec)

        n = len(df)
        ages = np.zeros(n)
        curage = 0
        i=0
        for line in open(filename):
            m = re.match('#',line)
            if m:
                m = re.match('#AGE=\s*(\d+\.\d+)\s+',line)
                if m:
                    curage=float(m.group(1))
            else:
                if re.search('\d',line):
                    ages[i]=curage
                    i+=1

        df['age'] = ages
        df['feh'] = cls.get_feh(filename)
        return df

    def df_all(self, phot):
        df = super(DartmouthModelGrid, self).df_all(phot)
        df.loc[:,'age'] = np.log10(df.age * 1e9) # convert to log10(age)
        df = df.sort_values(by=['feh','age','MMo','EEP'])
        df.index = [df.feh, df.age]
        return df

    def hdf_filename(self, phot):
        afe = self.kwargs['afe']
        y = self.kwargs['y']
        afe_str = '_{}'.format(afe) if afe!='afep0' else ''
        return os.path.join(self.datadir,'{}{}.h5'.format(phot, afe_str, y))



import os, os.path, re, sys
import logging
import time

from .config import on_rtd

if not on_rtd:
    import tables
    import matplotlib.pyplot as plt

from configobj import ConfigObj
from .starmodel import StarModel
from .isochrone import get_ichrone

def initLogging(filename, logger):
    if logger == None:
        logger = logging.getLogger()
    else:  # wish there was a logger.close()
        for handler in logger.handlers[:]:  # make a copy of the list
            logger.removeHandler(handler)

    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s: %(message)s')

    fh = logging.FileHandler(filename)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    sh = logging.StreamHandler(sys.stdout)
    logger.addHandler(sh)
    return logger

def starfit(folder, multiplicities=['single'], models='mist',
            use_emcee=False, plot_only=False, overwrite=False, verbose=False,
            logger=None, starmodel_type=None, ini_file='star.ini',
            no_plots=False, **kwargs):
    """ Runs starfit routine for a given folder.
    """
    nstars = {'single':1,
              'binary':2,
              'triple':3}

    if starmodel_type is None:
        Mod = StarModel
    else:
        Mod = starmodel_type

    ichrone = None

    print('Fitting {}'.format(folder))
    for mult in multiplicities:
        print('{} starfit...'.format(mult))
        model_filename = '{}_starmodel_{}.h5'.format(models, mult)

        #initialize logger for folder
        logfile = os.path.join(folder, 'starfit.log')
        logger = initLogging(logfile, logger)

        name = os.path.basename(os.path.abspath(folder))
        mod = None
        try:
            start = time.time()
            if plot_only:
                try:
                    mod = Mod.load_hdf('{}/{}'.format(folder,model_filename),
                                           name=name)
                except:
                    pass
            else:
                # Only try to fit model if it doesn't exist, unless overwrite is set
                fit_model = True

                try:
                    mod = Mod.load_hdf('{}/{}'.format(folder,model_filename),
                                         name=name)
                    fit_model = False
                except:
                    filename = '{}/{}'.format(folder,model_filename)
                    if os.path.exists(filename):
                        os.remove(filename)

                if fit_model or overwrite:
                    ini_file_path = os.path.join(folder, ini_file)
                    c = ConfigObj(ini_file_path)

                    if ichrone is None:
                        bands = StarModel.get_bands(ini_file_path)
                        ichrone = get_ichrone(models, bands)

                    N = nstars[mult] if 'N' not in c else None
                    mod = Mod.from_ini(ichrone, folder, use_emcee=use_emcee, N=N,
                                        ini_file=ini_file, name=name)
                    try:
                        mod.obs.print_ascii()
                    except:
                        pass

                    # # Prime the jit call signature?
                    # ichrone.mag['J'](1.0, 9.7, 0.1, 1000, 0.4)

                    mod.fit(verbose=verbose, overwrite=overwrite, **kwargs)
                    mod.save_hdf(os.path.join(folder, model_filename), overwrite=overwrite)
                else:
                    logger.info('{} exists.  Use -o to overwrite.'.format(model_filename))

            # Only make corner plots if they are older
            #  than the starmodel hdf file
            make_corners = False
            for x in ['physical', 'observed']:
                f = os.path.join(folder,
                                 '{}_corner_{}_{}.png'.format(models, mult, x))
                if not os.path.exists(f):
                    make_corners = True
                    break
                else:
                    t_mod = os.path.getmtime(os.path.join(folder,model_filename))
                    t_plot = os.path.getmtime(f)
                    if t_mod > t_plot:
                        make_corners=True

            if (make_corners or plot_only) and not no_plots:
                corner_base = os.path.join(folder, '{}_corner_{}'.format(models, mult))
                fig1,fig2 = mod.corner_plots(corner_base)


            # Make mag plot if necessary.
            magplot_file = os.path.join(folder, '{}_mags_{}.png'.format(models, mult))
            make_magplot = False #True
            if os.path.exists(magplot_file):
                if os.path.getmtime(os.path.join(folder, model_filename)) > \
                        os.path.getmtime(magplot_file) or \
                        plot_only:
                    pass
                else:
                    make_magplot = False

            if make_magplot:
                fig = mod.mag_plot()
                plt.savefig(os.path.join(folder,'{}_mags_{}.png'.format(models, mult)))

            end = time.time()
            if plot_only:
                logger.info('{} starfit successful (plots only) for '.format(mult) +
                            '{} in {:.1f} minutes.'.format(folder, (end-start)/60))
            else:
                logger.info('{} starfit successful for '.format(mult) +
                            '{} in {:.1f} minutes.'.format(folder, (end-start)/60))
        except KeyboardInterrupt:
            logger.error('{} starfit calculation interrupted for {}.'.format(mult,folder))
            raise
        except:
            logger.error('{} starfit calculation failed for {}.'.format(mult,folder),
                         exc_info=True)

    return mod, logger



from __future__ import division,print_function
import os,os.path
import numpy as np
import pkg_resources

from scipy.interpolate import LinearNDInterpolator as interpnd
import pandas as pd
import pickle

from .isochrone import Isochrone

#DATADIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))
DATADIR = os.path.expanduser('~/.isochrones')
if not os.path.exists(DATADIR):
    os.mkdir(DATADIR)

MASTERFILE = '{}/padova.h5'.format(DATADIR)
TRI_FILE = '{}/padova.tri'.format(DATADIR)

def _download_h5():
    url = 'http://zenodo.org/record/12800/files/padova.h5'
    from six.moves import urllib
    print('Downloading Padova stellar model data (should happen only once)...')
    if os.path.exists(MASTERFILE):
        os.remove(MASTERFILE)
    urllib.request.urlretrieve(url,MASTERFILE)

def _download_tri():
    url = 'http://zenodo.org/record/12800/files/padova.tri'
    from six.moves import urllib
    print('Downloading Padova isochrone pre-computed triangulation (should happen only once...)')
    if os.path.exists(TRI_FILE):
        os.remove(TRI_FILE)
    urllib.request.urlretrieve(url,TRI_FILE)

if not os.path.exists(MASTERFILE):
    _download_h5()

if not os.path.exists(TRI_FILE):
    _download_tri()

MASTERDF = pd.read_hdf(MASTERFILE,'df')


class Padova_Isochrone(Isochrone):
    """Padova Stellar Models
    """
    def __init__(self,bands=['bol','Kepler','g','r','i',
                             'z','D51','J','H','K']):

        df = MASTERDF

        mags = {}
        for band in bands:
            mags[band] = df[band]

        try:
            f = open(TRI_FILE,'rb')
            tri = pickle.load(f)
        except:
            f = open(TRI_FILE,'rb')
            tri = pickle.load(f,encoding='latin-1')
        finally:
            f.close()
        
        Isochrone.__init__(self,df['M_ini'],df['age'],df['feh'],
                           df['M_act'],df['logL'],
                           10**df['logTeff'],df['logg'],mags,tri=tri)

    
######## setup functions.  shouldn't have to use ever.

DATAFOLDER = '' #datafolder where raw padova tracks would live in order to recreate dataframe, triangulation, etc.

def fehstr(feh,minfeh=-2.0,maxfeh=0.2):
        if feh < minfeh:
            return '%.1f' % minfeh
        elif feh > maxfeh:
            return '%.1f' % maxfeh
        elif (feh > -0.05 and feh < 0):
            return '0.0'
        else:
            return '%.1f' % feh            


def padova_to_df(save=True,savefile='isochrones/data/padova.h5'):
    alldf = pd.DataFrame()
    
    for feh in np.arange(-2,0.21,0.1):
        filename = DATAFOLDER + '/padova_%s.dat' % fehstr(feh,-2,0.2)    
        age,m_ini,m_act,logL,logT,logg,mbol,kep,g,r,i,z,dd051,J,H,K = \
                np.loadtxt(filename,usecols=(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15),unpack=True)
        df = pd.DataFrame()
        df['age'] = age
        df['M_ini'] = m_ini
        df['M_act'] = m_act
        df['logL'] = logL
        df['logTeff'] = logT
        df['logg'] = logg
        df['bol'] = mbol
        df['Kepler'] = kep
        df['g'] = g
        df['r'] = r
        df['i'] = i
        df['z'] = z
        df['D51'] = dd051
        df['J'] = J
        df['H'] = H
        df['K'] = K
        df['feh'] = feh * np.ones_like(age)
        alldf = alldf.append(df)

    if save:
        alldf.to_hdf(savefile,'df')
        
    return alldf

def write_tri(df=MASTERDF,outfile=TRI_FILE):
        N = len(df)
        pts = np.zeros((N,3))
        pts[:,0] = np.array(df['M_ini'])
        pts[:,1] = np.array(df['age'])
        pts[:,2] = np.array(df['feh'])
        Jmags = np.array(df['J'])

        Jfn = interpnd(pts,Jmags)

        f = open(outfile,'wb')
        pickle.dump(Jfn.tri,f)
        f.close()



from isochrones.grid import ModelGrid
from isochrones.isochrone import Isochrone
import os, re, glob
import numpy as np
import pandas as pd
from isochrones.config import ISOCHRONES


class YAPSIModelGrid(ModelGrid):
    
    datadir = os.path.expanduser('~/yapsi')
    default_bands = list('UBVRIJHK')
    default_kwargs = {'Y':0.28}
    
    @classmethod
    def get_band(cls, b):
        if b in 'UBVRIJHK':
            return 'w', b
        else:
            raise ValueError('{0} not in YAPSI grids.'.format(b))
            
    @classmethod
    def _get_XYZ(cls, filename):
        m = re.search('X(\dp\d+)_Z(\dp\d+)\.dat', filename)
        if m:
            X = float(m.group(1).replace('p', '.'))
            Z = float(m.group(2).replace('p', '.'))
            Y = 1 - X - Z
            return X, Y, Z
        else:
            raise ValueError('Cannot parse XYZ from filename: {}'.format(filename))        

    def get_filenames(self, phot='w', Y=0.28):
        all_files = glob.glob('{0}/yapsi_{1}_*.dat'.format(self.datadir, phot))
        files = []
        for f in all_files:
            x, y, Z = self._get_XYZ(f)
            if np.isclose(Y, y):
                files.append(f)

        return files


    @classmethod
    def get_feh(cls, filename):
        """
        example filename: yapsi_w_X0p602357_Z0p027643.dat
        """
        X,Y,Z = cls._get_XYZ(filename)

        Xsun = 0.703812
        Zsun = 0.016188

        return np.log10((Z/X) / (Zsun/Xsun))
        
    @classmethod
    def to_df(cls, filename):

        df = pd.read_table(filename, comment='#',
             names=['age','mass','logTeff','logL','logg','V','UmB','BmV','VmR','VmI','JmK','HmK','VmK'],
             delim_whitespace=True)
        
        df['age'] = np.log10(df['age']*1e9)
        df['feh'] = cls.get_feh(filename)
        df['B'] = df['BmV'] + df['V']
        df['U'] = df['UmB'] + df['B']
        df['R'] = df['V'] - df['VmR']
        df['I'] = df['V'] - df['VmI']
        df['K'] = df['V'] - df['VmK']
        df['H'] = df['K'] - df['HmK']
        df['J'] = df['K'] - df['JmK']
        
        df = df.drop(['UmB','BmV','VmR','VmI','JmK','HmK','VmK'], axis=1)
        return df

    def _get_df(self):
        return self.df_all('w')
    


import os
import numpy as np

from isochrones.dartmouth import Dartmouth_Isochrone
from isochrones.mist import MIST_Isochrone
from isochrones import StarModel, BinaryStarModel, TripleStarModel

FOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__)))

DAR = Dartmouth_Isochrone()
MIST = MIST_Isochrone()

def test_ini():
    _check_ini(DAR)
    _check_ini(MIST)


#################

def _check_ini(ic):
    single_dirs = ['star1']
    binary_dirs = ['star2']
    triple_dirs = ['star3', 'star4']

    for d in single_dirs:
        SingleCheck().check(ic, os.path.join(FOLDER, d))
    for d in binary_dirs:
        BinaryCheck().check(ic, os.path.join(FOLDER, d))
        BinaryCheck_Unassoc().check(ic, os.path.join(FOLDER, d))
    for d in triple_dirs:
        TripleCheck().check(ic, os.path.join(FOLDER, d))
        TripleCheck_Unassoc1().check(ic, os.path.join(FOLDER, d))
        TripleCheck_Unassoc2().check(ic, os.path.join(FOLDER, d))

    # _ini1(ic)
    # _ini2(ic)
    # _ini3(ic)
    # _ini3_2(ic)
    # _ini4(ic)

class IniCheck(object):
    index = 0

    def get_mod(self, ic, folder):
        return StarModel.from_ini(ic, folder=folder, index=self.index)

    def check_asserts(self, mod):
        assert self.n_params == len(self.pars)
        assert mod.n_params == self.n_params
        assert mod.obs.systems == self.systems
        assert mod.obs.Nstars == self.Nstars
        assert np.isfinite(mod.lnlike(self.pars))

    def check_p0(self, mod):
        p0 = mod.emcee_p0(200)
        nbad = 0
        for i,p in enumerate(p0):
            if not np.isfinite(mod.lnpost(p)):
                print(p)
                nbad += 1
        assert nbad==0        

    def check(self, ic, folder):
        mod = self.get_mod(ic, folder)
        self.check_asserts(mod)
        self.check_p0(mod)

        if hasattr(self, 'get_mod_special'):
            mod = self.get_mod_special(ic, folder)
            self.check_asserts(mod)

class SingleCheck(IniCheck):
    pars = [1.0, 9.4, 0.0, 100, 0.2]
    n_params = 5
    systems = [0]
    Nstars = {0:1}

class BinaryCheck(IniCheck):
    pars = [1.0, 0.5, 9.4, 0.0, 100, 0.2]
    n_params = 6
    systems = [0]
    Nstars = {0:2}

    def get_mod_special(self, ic, folder):
        return BinaryStarModel(ic, folder=folder)

class BinaryCheck_Unassoc(IniCheck):
    pars = [1.0, 9.4, 0.0, 100, 0.2, 
            0.8, 9.7, 0.1, 300, 0.3]
    n_params = 10
    index = [0, 1]
    systems = [0, 1]
    Nstars = {0:1, 1:1}

class TripleCheck(IniCheck):
    pars = [1.0, 0.8, 0.5, 9.4, 0.0, 100, 0.2]
    n_params = 7
    systems = [0]
    Nstars = {0:3}

    def get_mod_special(self, ic, folder):
        return TripleStarModel(ic, folder=folder)


class TripleCheck_Unassoc1(IniCheck):
    pars = [1.0, 0.8, 9.4, 0.0, 100, 0.2, 
            1.0, 9.7, 0.0, 200, 0.5]
    n_params = 11
    index = [0, 0, 1]
    systems = [0, 1]
    Nstars = {0:2, 1:1}

class TripleCheck_Unassoc2(IniCheck):
    pars = [1.0, 9.4, 0.0, 100, 0.2, 
            1.0, 0.8, 9.7, 0.0, 200, 0.5]
    n_params = 11
    index = [0, 1, 1]
    systems = [0, 1]
    Nstars = {0:1, 1:2}


